from __future__ import absolute_import

import pandas as pd
import numpy as np
import os
import calendar
import logging

from xlrd.biffh import XLRDError


def read_ercot_da_spp(fname, month, settlement_point):
    """
    Reads the day-ahead market historical settlement point prices file at fname and returns the NumPy ndarray corresponding to the hourly price at settlement_point.

    :param fname: string giving location of DAM SPPs file
    :type fname: str
    :param month: which month to read data for; (int) [1, 12] OR (str) ['1', '12']
    :type month: int or str
    :param settlement_point: string giving settlement point name (hub or load zone)
    :type settlement_point: str
    :return spp_da: NumPy ndarray with SPPs for month and settlement point
    :rtype: NumPy ndarray
    """
    spp_da = np.array([])

    # if month is provided as an int, map it to the correct calendar month
    if isinstance(month, int):
        month_abbr = calendar.month_abbr(month)
    elif isinstance(month, str):
        month_ix = int(month)        
        month_abbr = calendar.month_abbr[month_ix]

    # Retrieve the correct worksheet for the month.
    wkbk = pd.ExcelFile(fname)
    wkbk_sheetnames = [name[:3] for name in wkbk.sheet_names]

    try:
        wkst_ix = wkbk_sheetnames.index(month_abbr)
    except ValueError:
        # The worksheet for the requested month does not exist.
        logging.warning('read_ercot_da_spp: Could not load data (the specified month of data could not be found in the given file), returning empty array. (got {fname}, {month}, {settlement_point})'.format(fname=fname, month=month, settlement_point=settlement_point))
        return spp_da
    else:
        df = wkbk.parse(wkst_ix)

        # filter DataFrame by settlement_point and extract series
        df0 = df.loc[df['Settlement Point'] == settlement_point]
        df1 = df0['Settlement Point Price']

        # convert to NumPy array and remove NaN
        spp_da = df1.astype('float').values
        spp_da = spp_da[~np.isnan(spp_da)]

    return spp_da


def read_ercot_da_ccp(fname, month):
    """
    Reads the day-ahead market historical capacity clearing prices file at fname and returns NumPy ndarrays corresponding to the hourly regdn and regup prices.

    :param fname: 
    :param month: which month to read data for; (int) [1, 12] OR (str) ['1', '12']
    
    :param fname: string giving location of DAM CCPs file
    :type fname: str
    :param month: which month to read data for; (int) [1, 12] OR (str) ['1', '12']
    :type month: int or str
    :return: NumPy ndarrays with regdn and regup CCPs for month
    :rtype: Numpy ndarrays
    """
    regdn = np.array([])
    regup = np.array([])

    # if month is provided as an int, map it to the correct calendar month
    if isinstance(month, int):
        month_ix = month
    elif isinstance(month, str):
        month_ix = int(month)
    
    # Read .csv and generate a Series for the month from "Delivery Date" column.
    df = pd.read_csv(fname, low_memory=False)
    series_month = pd.to_datetime(df['Delivery Date']).dt.month

    # Filter DataFrame by month.
    df1 = df.loc[series_month == month_ix]

    if len(df1) > 0:
        regdn = df1['REGDN'].astype('float').values
        regdn = regdn[~np.isnan(regdn)]

        try:
            regup = df1['REGUP '].astype('float').values  # why is there an extra space in the key
        except KeyError:
            regup = df1['REGUP'].astype('float').values
        finally:
            regup = regup[~np.isnan(regup)]
    else:
        logging.warning('read_ercot_da_ccp: No data matching input parameters found, returning empty array. (got {fname}, {month})'.format(fname=fname, month=month))
    
    return regdn, regup


def read_nodeid(fname,iso):
    from xlrd import open_workbook
    wb = open_workbook(filename = fname)
    ws = wb.sheet_by_name(iso)
    nodeid=[]
    for i in range(1,ws.nrows-1):
        cell=ws.cell(i,0)
        text=str(cell.value)
        text=text.replace('.0','')
        nodeid+=[text]
    return nodeid


def read_pjm_data(fpath, year, month, nodeid):
    """"
    Reads the historical LMP, regulation capacity, and regulation service (mileage) prices for the year 'year',
    the month 'month' and for the node 'nodeid'. Returns NumPy ndarrays for those three prices.

    :param fpath: The path to the root of the PJM data directory
    :type fpath: str
    :param year: Year of data to read
    :type year: int or str
    :param month: Month of data to read
    :type month: int or str
    :return: daLMP, RegCCP, RegPCP: Hourly LMP and regulation capacity/performance clearing price values.
    :rtype: NumPy ndarrays
    """
    daLMP = np.array([])
    RegCCP = np.array([])
    RegPCP = np.array([])
    rega = np.array([])
    regd = np.array([])
    mr = np.array([])

    if isinstance(month, str):
        month = int(month)
    
    if isinstance(year, str):
        year = int(year)
    
    if isinstance(nodeid, str):
        nodeid = int(nodeid)

    fnameLMP = "{0:d}{1:02d}_dalmp_{2:d}.csv".format(year, month, nodeid)
    fnameREG = "{0:d}{1:02d}_regp.csv".format(year,month)
    fnameMILEAGE = "{0:d}{1:02d}_regm.csv".format(year,month)

    fname_path_LMP = os.path.join(fpath, 'LMP', str(nodeid), str(year), fnameLMP)
    fname_path_REG = os.path.join(fpath, 'REG', str(year), fnameREG)
    fname_path_MILEAGE = os.path.join(fpath, 'MILEAGE', str(year), fnameMILEAGE)

    try:
        dfLMP = pd.read_csv(fname_path_LMP,index_col=False)
        daLMP = dfLMP['total_lmp_da'].values
    except FileNotFoundError:
        logging.warning('read_pjm_data: No LMP data matching input parameters found, returning empty array. (got {fname}, {year}, {month}, {nodeid})'.format(fname=fnameLMP, year=year, month=month, nodeid=nodeid))
    
    try:
        dfREG = pd.read_csv(fname_path_REG,index_col=False)
        RegCCP = dfREG['rmccp'].values
        RegPCP = dfREG['rmpcp'].values
    except FileNotFoundError:
        logging.warning('read_pjm_data: No REG data matching input parameters found, returning empty array. (got {fname}, {year}, {month})'.format(fname=fnameREG, year=year, month=month))

    try:
        dfMILEAGE = pd.read_csv(fname_path_MILEAGE,index_col=False)
        dfMILEAGE['MILEAGE RATIO'] = dfMILEAGE['regd_hourly']/dfMILEAGE['rega_hourly']

        # TODO: Handling NaNs/missing data intelligently. The current method just fills forward.
        rega = dfMILEAGE['rega_hourly'].replace([np.inf, -np.inf], np.nan).fillna(method='ffill').values
        regd = dfMILEAGE['regd_hourly'].replace([np.inf, -np.inf], np.nan).fillna(method='ffill').values
        mr = dfMILEAGE['MILEAGE RATIO'].replace([np.inf, -np.inf], np.nan).fillna(method='ffill').values
    except FileNotFoundError:
        logging.warning('read_pjm_data: No MILEAGE data matching input parameters found, returning empty array. (got {fname}, {year}, {month})'.format(fname=fnameMILEAGE, year=year, month=month))

    return daLMP, mr, rega, regd, RegCCP, RegPCP


def read_pjm_da_lmp(fname, node_name):
    """
    Reads the day-ahead LMP file at fname and returns the NumPy ndarray corresponding to the hourly LMP at node_name.
    :param fname: A string containing the path to the relevant day-ahead LMP file.
    :param node_name: A string containing the name of the pricing node of interest.
    :return: LMP: A NumPy ndarray containing the hourly LMP at node-name.
    """
    # read in the .csv file
    df = pd.read_csv(fname, low_memory=False)

    # filter rows by node_name
    col2 = df.axes[1][2]
    pnode_ix = df.index[df[col2] == node_name]
    df1 = df.iloc[pnode_ix, :]

    # filter Total LMP columns
    df2 = df1[df1.axes[1][7:79:3]]

    # convert to NumPy ndarray, ravel, and remove NaNs
    LMP = np.ravel(df2.astype('float').values)
    LMP = LMP[~np.isnan(LMP)]

    return LMP


def read_pjm_reg_signal(fname):
    """
    Reads the regulation signal file at fname and returns the NumPy ndarray corresponding to the hourly integrated signal.
    :param fname: A string containing the path to the relevant regulation signal file.
    :return: RU, RD: NumPy ndarrays containing the hourly integrated regulation up/down signals.
    """
    # read in the Excel file
    df = pd.read_excel(fname, skip_footer=1)

    # create DateTime indexing to facilitate resampling
    dt_ix = pd.date_range('2017-11-01', periods=30*60*24, freq='2S')
    df.index = dt_ix

    # define function for performing hourly integration
    def _hourly_integration(array_like):
        # ZOH integration
        dt = 2.0/(60*60)
        return np.sum(array_like)*dt

    # use resample to apply hourly integration
    df1 = df.resample(rule='H', closed='left').apply(_hourly_integration)

    # convert DataFrame to NumPy ndarray and ravel
    REG = np.ravel(df1.astype('float').values, 'F')

    # assign reg up/down values appropriately based on sign of regulation signal
    RU = REG * (REG >= 0)
    RD = REG * (REG < 0)

    return RU, RD


def read_pjm_mileage(fname, month):
    """
    Reads the historic regulation market data file at fname and returns the NumPy ndarrays for mileage data.
    :param fname: A string containing the path to the relevant historic regulation market data file.
    :param month: An int corresponding to the month of interest (1: Jan., 2: Feb., etc.)
    :return: Mileage_Ratio, RegA_milage, RegD_mileage: NumPy ndarrays containing computed mileage ratio and RegA/RegD hourly mileage signals.
    """
    # read in the Excel file and parse the relevant worksheet
    wkbk = pd.ExcelFile(fname)
    df = wkbk.parse(month)

    # replace "UNAPPROVED" entries with previous filled value
    df.fillna(method='ffill', inplace=True)

    # compute hourly mileage ratio
    # TODO: what do infinite mileage ratio? (REGA = 0)
    df['MILEAGE RATIO'] = df['REGD_HOURLY'] / df['REGA_HOURLY']

    RegA_mileage = df['REGA_HOURLY'].values
    RegD_mileage = df['REGD_HOURLY'].values
    Mileage_Ratio = df['MILEAGE RATIO'].values

    return Mileage_Ratio, RegA_mileage, RegD_mileage


def read_pjm_reg_price(fname, month):
    """
    Reads the historical ancillary services data file at fname and returns the NumPy ndarrays for regulation clearing prices.
    :param fname: A string containing the path to the relevant historical ancillary services data file.
    :param month: An int corresponding to the month of interest (1: Jan., 2: Feb., etc.)
    :return: RegCCP, RegPCP: NumPy ndarrays containing hourly regulation capacity/performance clearing price values.
    """
    # read in the Excel file and parse relevant worksheet
    wkbk = pd.ExcelFile(fname)
    df = wkbk.parse(month)

    # parse the relevant service
    df1 = df[df['SERVICE'] == 'REG']

    # DLS hour gives NaN
    df1.fillna(method='ffill', inplace=True)

    RegCCP = df1['REG_CCP'].values
    RegPCP = df1['REG_PCP'].values

    return RegCCP, RegPCP


def read_miso_da_lmp(fname, node_name):
    """
    Reads the day-ahead LMP file at fname and returns the NumPy ndarray corresponding to the hourly LMP at node_name.
    :param fname: A string containing the path to the relevant day-ahead LMP file.
    :param node_name: A string containing the name of the pricing node of interest.
    :return: LMP: A NumPy ndarray containing the hourly LMP at node-name.
    """
    # parse fname for month and year values
    month = int(fname[-2:])
    year = int(fname[-6:-2])

    nday = {1: 31, 2: 28, 3: 31, 4: 30, 5: 31, 6: 30, 7: 31, 8: 31, 9: 30, 10: 31, 11: 30, 12: 31}
    if year%4 == 0:
        nday[2] = 29

    LMP = np.empty([0])
    for day in range(1, nday[month]+1):
    
        if (year <= 2014) or (year == 2015 and month <= 2):
            fname_ = fname + str(day).zfill(2) + "_da_lmp.csv"
        else: 
            fname_ = fname + str(day).zfill(2) + "_da_exante_lmp.csv"
        
        df = pd.read_csv(fname_, skiprows=4, low_memory=False)
        # filter rows by node_name
        col1 = df.axes[1][0]
        col3 = df.axes[1][2]
        pnode_ix1 = df.index[df[col1] == node_name]
        df1 = df.iloc[pnode_ix1, :]
        
        # find LMP values
        pnode_ix2 = df1.index[df1[col3] == "LMP"]
        df2 = df.iloc[pnode_ix2, :]
        
        # filter Total LMP columns
        df3 = df2[df2.axes[1][3:27]]

        # convert to NumPy ndarray, ravel, and remove NaNs
        LMP_day = np.ravel(df3.astype('float').values)
        LMP_day = LMP_day[~np.isnan(LMP_day)] 
        LMP = np.append(LMP, LMP_day)
    
    return LMP


def read_miso_reg_price(fname):
    """
    Reads the historical ancillary services data file at fname and returns the NumPy ndarrays for regulation clearing prices.
    :param fname: A string containing the path to the relevant historical ancillary services data file.
    :return: RegMCP: NumPy ndarrays containing hourly regulation capacity/mileage clearing price values.
    """
    # parse fname for month and year values
    month = int(fname[-2:])
    year = int(fname[-6:-2])

    nday = {1: 31, 2: 28, 3: 31, 4: 30, 5: 31, 6: 30, 7: 31, 8: 31, 9: 30, 10: 31, 11: 30, 12: 31}
    if year%4 == 0:
        nday[2] = 29

    RegMCP = np.empty([0])

    for day in range(1, nday[month] + 1):
        if (year <= 2014) or (year == 2015 and month <= 2):
            fname_ = fname + str(day).zfill(2) + "_asm_damcp.csv"
        else: 
            fname_ = fname + str(day).zfill(2) + "_asm_exante_damcp.csv"

        # read in the .csv file
        df = pd.read_csv(fname_, skiprows=4, nrows=7, low_memory=False)
        
        # find SERREGMCP values
        col3 = df.axes[1][2]
        pnode_ix1 = df.index[df[col3] == "SERREGMCP"]
        df1 = df.iloc[pnode_ix1, :]
        df2 = df1[df1.axes[1][3:27]]
    
        # convert to NumPy ndarray, ravel, and remove NaNs
        RegMCP_day = np.ravel(df2.astype('float').values)
        RegMCP_day = RegMCP_day[~np.isnan(RegMCP_day)] 
        RegMCP = np.append(RegMCP, RegMCP_day)
    
    return RegMCP

##################################################################################################################
#///////////////////////////////////////////////////////#
def read_isone_data(fpath, year, month, nodeid):
    """"
    Reads the historical LMP, regulation capacity, and regulation service (mileage) prices for the year 'year',
    the month 'month' and for the node 'nodeid'. Returns NumPy ndarrays for those three prices.
    :param fpath: A string containing the path to the relevant historical ancillary services data file.
    :param year: An int corresponding to the year of interest
    :param month: An int corresponding to the month of interest (1: Jan., 2: Feb., etc.)
    :return: daLMP, RegCCP, RegPCP: NumPy ndarrays containing hourly LMP as well as regulation capacity/performance clearing price values.
    """

    if isinstance(month, str):
        month = int(month)

    if isinstance(year, str):
        year = int(year)

    if isinstance(nodeid, (int, float, complex)):
        nodeid = str(nodeid)

    if year < 2018:
        if year == 2017 and month == 12:
            fnameLMP = "{0:d}{1:02d}_fmlmp_{2:s}.csv".format(year, month, nodeid)
            fnameRCP = "{0:d}{1:02d}_fmrcp.csv".format(year ,month)
        else:
            fnameLMP = "{0:d}{1:02d}_dalmp_{2:s}.csv".format(year, month, nodeid)
            fnameRCP = "{0:d}{1:02d}_rcp.csv".format(year ,month)
    else:
        fnameLMP = "{0:d}{1:02d}_fmlmp_{2:s}.csv".format(year, month, nodeid)
        fnameRCP = "{0:d}{1:02d}_fmrcp.csv".format(year ,month)

    fname_path_LMP = os.path.join(fpath, 'LMP', str(nodeid), str(year), fnameLMP)
    fname_path_RCP = os.path.join(fpath, 'RCP', str(year), fnameRCP)
    fname_path_MILEAGE = os.path.join(fpath, 'MileageFile.xlsx')

    daLMP = np.empty([0])
    RegCCP = np.empty([0])
    RegPCP = np.empty([0])
    miMULT = np.empty([0])


    try:
        dfLMP = pd.read_csv(fname_path_LMP, index_col=False)
        daLMP = dfLMP['LmpTotal'].values
    except FileNotFoundError:
        logging.warning \
            ('read_isone_data: No LMP data matching input parameters found, returning empty array. (got {fname}, {year}, {month}, {nodeid})'.format
                (fname=fnameLMP, year=year, month=month, nodeid=nodeid))

    try:
        if year > 2014:
            if year == 2015 and month < 4:
                dfRCP = pd.read_csv(fname_path_RCP, index_col=False)
                RegCCP = dfRCP['RegClearingPrice'].values
                RegPCP = []
            else:
                dfRCP = pd.read_csv(fname_path_RCP, index_col=False)
                RegCCP = dfRCP['RegCapacityClearingPrice'].values
                RegPCP = dfRCP['RegServiceClearingPrice'].values
                                   
                dataF_mileage_file = pd.read_excel(fname_path_MILEAGE, sheet_name = 'Energy Neutral Trinary', usecols = ['Fleet ATRR dispatch [MW]'])
                dataF_mileage_file = dataF_mileage_file.append(pd.DataFrame([-10]*15, columns = ['Fleet ATRR dispatch [MW]']), ignore_index = True) #   changes number of data points to 24 hours; doesn't change mileage
                            
                #   AGC setpoints given every 4 seconds, take the total mileage for each hour; total of one day of mileage
                hours = [i for i in range(len(dataF_mileage_file.index)//900)]
                
                mileage_day = []
                for hour in hours:
                    dataF_mileage_hour = dataF_mileage_file[900*hour:900*(hour + 1)] #    every 900 values represents an hour (900*4 = 3600)
                    mileage_hour = 0 
                    for i in range(len(dataF_mileage_hour.index)):
                        if i == len(dataF_mileage_hour.index) - 1:
                            break
                    
                        if not dataF_mileage_hour.iloc[i, 0] == dataF_mileage_hour.iloc[i + 1, 0]:
                            mileage_hour += abs(dataF_mileage_hour.iloc[i, 0] - dataF_mileage_hour.iloc[i + 1, 0])/10
                            
                    mileage_day.append(mileage_hour)
            
                dataF_mileage_day = pd.DataFrame(mileage_day, columns = ['Trinary Mileage'])
                
                #   have one days worth of data, need one months worth
                days = len(daLMP)//24
                mileage_mult = pd.DataFrame(columns = ['Trinary Mileage'])
                for day in range(days):
                    mileage_mult = mileage_mult.append(dataF_mileage_day, ignore_index = True)
                #   if the len are offset, make them match
                if not len(daLMP) == len(mileage_mult):
                    diff = len(daLMP) - len(mileage_mult)
                    
                    for i in range(diff):
                        mileage_mult = mileage_mult.append(dataF_mileage_day.iloc[i], ignore_index = True)
                        
                miMULT = mileage_mult['Trinary Mileage'].values
        else:
            dfRCP = pd.read_csv(fname_path_RCP, index_col=False)
            RegCCP = dfRCP['RegClearingPrice'].values
            RegPCP = []

    except FileNotFoundError:
        logging.warning \
            ('read_isone_data: No ASP data matching input parameters found, returning empty array. (got {fname}, {year}, {month})'.format
                (fname=fnameRCP, year=year, month=month))
    

    return daLMP, RegCCP, RegPCP, miMULT
#///////////////////////////////////////////////////////#


def read_miso_data(fpath, year, month, nodeid):
    """Reads the daily MISO data files and returns the NumPy ndarrays for LMP and MCP.

    
    :param fpath: root of the MISO data folder
    :type fpath: str
    :param year: year of data
    :type year: int or str
    :param month: month of data
    :type month: int or str
    :param nodeid: pricing node ID
    :type nodeid: str
    :return: arrays of data specified
    :rtype: NumPy ndarrays
    """
    LMP = np.array([])
    RegMCP = np.array([])

    _, n_days_month = calendar.monthrange(int(year), int(month))

    for day in range(1, n_days_month+1):
        # Read daily files.
        date_str = '{year}{month}{day}'.format(year=year, month=str(month).zfill(2), day=str(day).zfill(2))

        if (int(year) <= 2014) or (int(year) == 2015 and int(month) <= 2):
            lmp_fname = os.path.join(fpath, 'LMP', str(year), str(month).zfill(2), '{prefix}_da_lmp.csv'.format(prefix=date_str))
            mcp_fname = os.path.join(fpath, 'MCP', str(year), str(month).zfill(2), '{prefix}_asm_damcp.csv'.format(prefix=date_str))
        else: 
            lmp_fname = os.path.join(fpath, 'LMP', str(year), str(month).zfill(2), '{prefix}_da_exante_lmp.csv'.format(prefix=date_str))
            mcp_fname = os.path.join(fpath, 'MCP', str(year), str(month).zfill(2), '{prefix}_asm_exante_damcp.csv'.format(prefix=date_str))
        
        # LMP file.
        try:
            df = pd.read_csv(lmp_fname, skiprows=4, low_memory=False)
        except FileNotFoundError:
            logging.warning('read_miso_data: LMP file missing, returning empty array.')
            break

        # Filter rows by node_name.
        col1 = df.axes[1][0]
        col3 = df.axes[1][2]
        pnode_ix1 = df.index[df[col1] == nodeid]
        df1 = df.iloc[pnode_ix1, :]
        
        # Find LMP values.
        pnode_ix2 = df1.index[df1[col3] == "LMP"]
        df2 = df.iloc[pnode_ix2, :]
        
        # Filter Total LMP columns.
        df3 = df2[df2.axes[1][3:27]]

        if len(df3) == 0:
            LMP = np.array([])
            logging.warning('read_miso_data: A daily LMP file is missing required data, returning empty array.')
            break

        # Convert to NumPy ndarray, ravel, and remove NaNs.
        LMP_day = np.ravel(df3.astype('float').values)
        LMP_day = LMP_day[~np.isnan(LMP_day)] 
        LMP = np.append(LMP, LMP_day)

        # MCP file.
        try:
            df = pd.read_csv(mcp_fname, skiprows=4, nrows=7, low_memory=False)
        except FileNotFoundError:
            RegMCP = np.array([])
            logging.warning('read_miso_data: MCP file missing, returning empty array.')
            break
        
        # Find SERREGMCP values.
        col3 = df.axes[1][2]
        pnode_ix1 = df.index[df[col3] == "SERREGMCP"]
        df1 = df.iloc[pnode_ix1, :]
        df2 = df1[df1.axes[1][3:27]]
    
        # convert to NumPy ndarray, ravel, and remove NaNs
        RegMCP_day = np.ravel(df2.astype('float').values)
        RegMCP_day = RegMCP_day[~np.isnan(RegMCP_day)] 
        RegMCP = np.append(RegMCP, RegMCP_day)
    
    return LMP, RegMCP




#######################################################################################################################
# NYISO
#######################################################################################################################

def read_nyiso_data(fpath, year, month, nodeid, typedat="both", RT_DAM="both"):
    """"
    Reads the historical LBMP, regulation capacity, and regulation movement prices for the year 'year',
    the month 'month' and for the node 'nodeid'. Returns NumPy ndarrays for those three prices.

    :param fpath: The path to the root of the NYISO data directory
    :type fpath: str
    :param year: Year of data to read
    :type year: int or str
    :param month: Month of data to read
    :type month: int or str
    :param nodeid: ID of the node to read
    :type nodeid: int or str
    :return: daLBMP, rtLBMP, daCAP, rtCAP, rtMOV: Hourly LBMP and regulation capacity/movement clearing price values.
    :rtype: NumPy ndarraysx
    """
    ############################################################

    daLBMP = np.empty([0])
    rtLBMP = np.empty([0])
    daCAP = np.empty([0])
    rtCAP = np.empty([0])
    rtMOV = np.empty([0])

    if isinstance(month, str):
        month = int(month)

    if isinstance(year, str):
        year = int(year)

    if isinstance(nodeid, str):
        nodeid = int(nodeid)

    ############################################################################################
    # folderfile = fpath
    # # TODO: path_nodes_file is a folder to adjust when integrating it to QuESt
    # path_nodes_file = 'C:/Users/fwilche/Documents/data_bank/NYISO/'
    path_nodes_file = '../../es_gui/apps/data_manager/_static/'
    pathf_nodeszones = os.path.join(fpath, path_nodes_file, 'nodes_nyiso.csv')
    df_nodeszones = pd.read_csv(pathf_nodeszones, index_col=False)
    df_nodeszones_x = df_nodeszones.loc[df_nodeszones['Node ID'] == nodeid, :]

    if df_nodeszones_x.empty:
        print("The node does NOT exists in NYISO")
        # raise ValueError('Not a valid bus number!!!')
        return daLBMP, rtLBMP, daCAP, rtCAP, rtMOV
    else:
        if df_nodeszones_x.iloc[0,0] == df_nodeszones_x.iloc[0,2]:
            print("It's a zone node")
            zoneid = nodeid
            zone_gen = "zone"
        else:
            print("It's a gen node")
            zoneid = df_nodeszones_x.iloc[0,2]
            zone_gen = "gen"
    print("Identified zone:")
    print(zoneid)
    ############################################################################################

    ndaysmonth = calendar.monthrange(year, month)
    ndaysmonth = int(ndaysmonth[1])


    for ix in range(ndaysmonth):
        day_x = ix+1
        date_str = str(year)+ str(month).zfill(2)+str(day_x).zfill(2)
        # print(date_str)

        fnameLBMP_DA = date_str + "damlbmp_" + zone_gen + ".csv"
        fnameASP_DA = date_str + "damasp.csv"

        fnameLBMP_RT = date_str + "realtime_" + zone_gen + ".csv"
        fnameASP_RT = date_str + "rtasp.csv"

        fname_path_LBMP_DA = os.path.join(fpath, 'LBMP', 'DAM', zone_gen, str(year), str(month).zfill(2),fnameLBMP_DA)
        fname_path_ASP_DA = os.path.join(fpath, 'ASP', 'DAM', str(year), str(month).zfill(2),fnameASP_DA)

        fname_path_LBMP_RT = os.path.join(fpath, 'LBMP', 'RT', zone_gen, str(year), str(month).zfill(2),fnameLBMP_RT)
        fname_path_ASP_RT = os.path.join(fpath, 'ASP', 'RT', str(year), str(month).zfill(2),fnameASP_RT)

        if typedat == "asp" or typedat == "both":
            # 20170201damasp.csv
            # 20180501rtasp.csv
            if RT_DAM == "RT" or RT_DAM == "both":
                try:
                    df_file = pd.read_csv(fname_path_ASP_RT, index_col=False)
                except FileNotFoundError:
                    rtCAP = np.empty([0])
                    rtMOV = np.empty([0])
                    logging.warning('read_nyiso_data: RT ASP file missing, returning empty array.')
                    break

                if (year>=2016 and month>=6 and day_x>=23) or (year>=2016 and month>=7) or (year>=2017):
                    # NYCA Regulation Capacity ($/MWHr) - for newest type of data
                    df_file_rtCAP = df_file.loc[df_file['PTID'] == zoneid, ['NYCA Regulation Capacity ($/MWHr)']]
                    rtCAP = np.append(rtCAP, df_file_rtCAP.values)
                    # NYCA Regulation Movement ($/MW)
                    df_file_rtMOV = df_file.loc[df_file['PTID'] == zoneid, ['NYCA Regulation Movement ($/MW)']]
                    rtMOV = np.append(rtMOV, df_file_rtMOV.values)
                elif (year>=2001 and month>=10 and day_x>=0) or (year>=2001 and month>=11) or \
                     (year>=2002 and not ((year>=2016 and month>=6 and day_x>=23) or (year>=2016 and month>=7) or (year>=2017))):
                    df_file_rtCAP = df_file['East Regulation ($/MWHr)']
                    rtCAP = np.append(rtCAP, df_file_rtCAP.values)
                    df_file_rtMOV = df_file[' NYCA Regulation Movement ($/MW)']
                    rtMOV = np.append(rtMOV, df_file_rtMOV.values)
                    # RT ancillary services for NYISO start on July 2004


            if RT_DAM == "DAM" or RT_DAM == "both":
                try:
                    df_file = pd.read_csv(fname_path_ASP_DA, index_col=False)
                except FileNotFoundError:
                    daCAP = np.empty([0])
                    logging.warning('read_nyiso_data: DA ASP file missing, returning empty array.')
                    break

                if (year >= 2016 and month >= 6 and day_x >= 23) or (year >= 2016 and month >= 7) or (year >= 2017):
                    df_file_daCAP = df_file.loc[df_file['PTID'] == zoneid, ['NYCA Regulation Capacity ($/MWHr)']]
                    daCAP = np.append(daCAP, df_file_daCAP.values)
                elif (year >= 2001 and month >= 10 and day_x >= 0) or (year >= 2001 and month >= 11) or \
                     (year >= 2002 and not ((year >= 2016 and month >= 6 and day_x >= 23) or (year >= 2016 and month >= 7) or (year >= 2017))):
                    df_file_daCAP = df_file['East Regulation ($/MWHr)']
                    daCAP = np.append(daCAP, df_file_daCAP.values)
                else:
                    df_file_daCAP = df_file['Regulation ($/MWHr)']
                    daCAP = np.append(daCAP, df_file_daCAP.values)

        if typedat == "lbmp" or typedat == "both":
            # 20170201damlbmp_gen.csv
            # 20170201damlbmp_zone.csv
            # 20170201realtime_gen.csv
            if RT_DAM == "RT" or RT_DAM == "both":
                try:
                    df_rtLBMP = pd.read_csv(fname_path_LBMP_RT, index_col=False)
                except FileNotFoundError:
                    rtLBMP = np.empty([0])
                    logging.warning('read_nyiso_data: RT LMP file missing, returning empty array.')
                    break

                rtLBMP_node_x = df_rtLBMP.loc[df_rtLBMP['PTID'] == nodeid, ['LBMP ($/MWHr)']]
                rtLBMP = np.append(rtLBMP, rtLBMP_node_x.values)
                if rtLBMP_node_x.empty:
                    return np.empty([0]), np.empty([0]), np.empty([0]), np.empty([0]), np.empty([0])

            if RT_DAM == "DAM" or RT_DAM == "both":
                try:
                    df_daLBMP = pd.read_csv(fname_path_LBMP_DA, index_col=False)
                except FileNotFoundError:
                    daLBMP = np.empty([0])
                    logging.warning('read_nyiso_data: DA LMP file missing, returning empty array.')
                    break

                daLBMP_node_x = df_daLBMP.loc[df_daLBMP['PTID'] == nodeid, ['LBMP ($/MWHr)']]
                daLBMP = np.append(daLBMP, daLBMP_node_x.values)
                if daLBMP_node_x.empty:
                    return np.empty([0]), np.empty([0]), np.empty([0]), np.empty([0]), np.empty([0])

    return daLBMP, rtLBMP, daCAP, rtCAP, rtMOV


#TODO: delete function below:
def read_nyiso_data_old(fpath, year, month, nodeid, typedat="both", RT_DAM="both"):
    """"
    Reads the historical LBMP, regulation capacity, and regulation movement prices for the year 'year',
    the month 'month' and for the node 'nodeid'. Returns NumPy ndarrays for those three prices.

    :param fpath: The path to the root of the NYISO data directory
    :type fpath: str
    :param year: Year of data to read
    :type year: int or str
    :param month: Month of data to read
    :type month: int or str
    :param nodeid: ID of the node to read
    :type nodeid: int or str
    :return: daLBMP, rtLBMP, daCAP, rtCAP, rtMOV: Hourly LBMP and regulation capacity/movement clearing price values.
    :rtype: NumPy ndarraysx
    """
    ############################################################

    if isinstance(month, str):
        month = int(month)

    if isinstance(year, str):
        year = int(year)

    if isinstance(nodeid, str):
        nodeid = int(nodeid)

    # folderfile = fpath
    pathfile_aspzones = os.path.join(fpath, 'NYISO_aspnodes_list.csv')
    df_aspzones = pd.read_csv(pathfile_aspzones, index_col=False)

    df_aspzones_x = df_aspzones.loc[df_aspzones['PTID']== nodeid, :]

    if df_aspzones_x.empty:
        print("It's empty -must be a gen and not a zone node")

        # pathfile_genzones = folderfile + 'generator_NYISO.csv'
        pathfile_genzones = os.path.join(fpath, 'generator_NYISO.csv')
        df_genzones = pd.read_csv(pathfile_genzones, index_col=False)

        df_genzones_x = df_genzones.loc[df_genzones['PTID'] == nodeid, ['Zone']]

        if not df_genzones_x.empty:
            zone_gen = "gen"
            zone_x = df_genzones_x.iloc[0,0]

            df_zoneid_x = df_aspzones.loc[df_aspzones['Name'] == zone_x, 'PTID']

            if not df_zoneid_x.empty:
                print("Identified zone:")
                print(df_zoneid_x)
                zoneid = df_zoneid_x.iloc[0]
            else:
                raise ValueError('Bus outside NYISO!!!')

        else:
            raise ValueError('Not a valid bus number!!!')



    else:
        print("It's a zone node")
        zoneid = nodeid
        zone_gen = "zone"

    ############################################################

    ndaysmonth = calendar.monthrange(year, month)
    ndaysmonth = int(ndaysmonth[1])

    daLBMP = np.empty([0])
    rtLBMP = np.empty([0])
    daCAP = np.empty([0])
    rtCAP = np.empty([0])
    rtMOV = np.empty([0])

    for ix in range(ndaysmonth):
        day_x = ix+1
        date_str = str(year)+ str(month).zfill(2)+str(day_x).zfill(2)
        # print(date_str)

        fnameLBMP_DA = date_str + "damlbmp_" + zone_gen + ".csv"
        fnameASP_DA = date_str + "damasp.csv"

        fnameLBMP_RT = date_str + "realtime_" + zone_gen + ".csv"
        fnameASP_RT = date_str + "rtasp.csv"

        fname_path_LBMP_DA = os.path.join(fpath, 'LBMP', 'DAM', zone_gen, str(year), str(month).zfill(2),fnameLBMP_DA)
        fname_path_ASP_DA = os.path.join(fpath, 'ASP', 'DAM', str(year), str(month).zfill(2),fnameASP_DA)

        fname_path_LBMP_RT = os.path.join(fpath, 'LBMP', 'RT', zone_gen, str(year), str(month).zfill(2),fnameLBMP_RT)
        fname_path_ASP_RT = os.path.join(fpath, 'ASP', 'RT', str(year), str(month).zfill(2),fnameASP_RT)

        if typedat == "asp" or typedat == "both":
            # 20170201damasp.csv
            # 20180501rtasp.csv
            if RT_DAM == "RT" or RT_DAM == "both":
                try:
                    df_file = pd.read_csv(fname_path_ASP_RT, index_col=False)
                except FileNotFoundError:
                    rtCAP = np.empty([0])
                    rtMOV = np.empty([0])
                    logging.warning('read_nyiso_data: RT ASP file missing, returning empty array.')
                    break

                if (year>=2016 and month>=6 and day_x>=23) or (year>=2016 and month>=7) or (year>=2017):
                    # NYCA Regulation Capacity ($/MWHr) - for newest type of data
                    df_file_rtCAP = df_file.loc[df_file['PTID'] == zoneid, ['NYCA Regulation Capacity ($/MWHr)']]
                    rtCAP = np.append(rtCAP, df_file_rtCAP.values)
                    # NYCA Regulation Movement ($/MW)
                    df_file_rtMOV = df_file.loc[df_file['PTID'] == zoneid, ['NYCA Regulation Movement ($/MW)']]
                    rtMOV = np.append(rtMOV, df_file_rtMOV.values)
                elif (year>=2001 and month>=10 and day_x>=0) or (year>=2001 and month>=11) or \
                     (year>=2002 and not ((year>=2016 and month>=6 and day_x>=23) or (year>=2016 and month>=7) or (year>=2017))):
                    df_file_rtCAP = df_file['East Regulation ($/MWHr)']
                    rtCAP = np.append(rtCAP, df_file_rtCAP.values)
                    df_file_rtMOV = df_file[' NYCA Regulation Movement ($/MW)']
                    rtMOV = np.append(rtMOV, df_file_rtMOV.values)
                    # RT ancillary services for NYISO start on July 2004


            if RT_DAM == "DAM" or RT_DAM == "both":
                try:
                    df_file = pd.read_csv(fname_path_ASP_DA, index_col=False)
                except FileNotFoundError:
                    daCAP = np.empty([0])
                    logging.warning('read_nyiso_data: DA ASP file missing, returning empty array.')
                    break

                if (year >= 2016 and month >= 6 and day_x >= 23) or (year >= 2016 and month >= 7) or (year >= 2017):
                    df_file_daCAP = df_file.loc[df_file['PTID'] == zoneid, ['NYCA Regulation Capacity ($/MWHr)']]
                    daCAP = np.append(daCAP, df_file_daCAP.values)
                elif (year >= 2001 and month >= 10 and day_x >= 0) or (year >= 2001 and month >= 11) or \
                     (year >= 2002 and not ((year >= 2016 and month >= 6 and day_x >= 23) or (year >= 2016 and month >= 7) or (year >= 2017))):
                    df_file_daCAP = df_file['East Regulation ($/MWHr)']
                    daCAP = np.append(daCAP, df_file_daCAP.values)
                else:
                    df_file_daCAP = df_file['Regulation ($/MWHr)']
                    daCAP = np.append(daCAP, df_file_daCAP.values)

        if typedat == "lbmp" or typedat == "both":
            # 20170201damlbmp_gen.csv
            # 20170201damlbmp_zone.csv
            # 20170201realtime_gen.csv
            if RT_DAM == "RT" or RT_DAM == "both":
                try:
                    df_rtLBMP = pd.read_csv(fname_path_LBMP_RT, index_col=False)
                except FileNotFoundError:
                    rtLBMP = np.empty([0])
                    logging.warning('read_nyiso_data: RT LMP file missing, returning empty array.')
                    break

                rtLBMP_node_x = df_rtLBMP.loc[df_rtLBMP['PTID'] == nodeid, ['LBMP ($/MWHr)']]
                rtLBMP = np.append(rtLBMP, rtLBMP_node_x.values)


            if RT_DAM == "DAM" or RT_DAM == "both":
                try:
                    df_daLBMP = pd.read_csv(fname_path_LBMP_DA, index_col=False)
                except FileNotFoundError:
                    daLBMP = np.empty([0])
                    logging.warning('read_nyiso_data: DA LMP file missing, returning empty array.')
                    break

                daLBMP_node_x = df_daLBMP.loc[df_daLBMP['PTID'] == nodeid, ['LBMP ($/MWHr)']]
                daLBMP = np.append(daLBMP, daLBMP_node_x.values)

    return daLBMP, rtLBMP, daCAP, rtCAP, rtMOV

#######################################################################################################################

#######################################################################################################################
# SPP
#######################################################################################################################

def read_spp_data(fpath, year, month, node, typedat="both"):
    """"
    Reads the historical LMP, regulation capacity, and regulation service (mileage) prices for the year 'year',
    the month 'month' and for the node 'nodeid'. Returns NumPy ndarrays for those three prices.
    :param fpath: A string containing the path to the relevant historical ancillary services data file.
    :param year: An int corresponding to the year of interest
    :param month: An int corresponding to the month of interest (1: Jan., 2: Feb., etc.)
    :param node: A string with the name of the node in SPP
    :param typedat: xxxxxx xxxxxxx
    :param RT_DAM: xxxxxxxxx xxxx
    :return: daLMP, daMCPRU, daMCPRD: NumPy ndarrays containing hourly LMP as well as regulation up and down prices for the SPP market
    """

    daLMP = np.empty([0])
    daMCPRU = np.empty([0])
    daMCPRD = np.empty([0])

    if isinstance(month, str):
        month = int(month)

    if isinstance(year, str):
        year = int(year)

    ############################################################################################
    # TODO: path_nodes_file is a folder to adjust when integrating it to QuESt
    path_nodes_file = '../../es_gui/apps/data_manager/_static/'
    pathf_nodeszones = os.path.join(fpath, path_nodes_file, 'nodes_spp.csv')
    df_nodes = pd.read_csv(pathf_nodeszones, index_col=False, encoding="cp1252")
    df_nodes_x = df_nodes.loc[df_nodes['Node ID'] == node, :]

    if df_nodes_x.empty:
        print("The node does NOT exists in SPP")
        # raise ValueError('Not a valid bus number!!!')
        return daLMP, daMCPRU, daMCPRD
    else:
        nodetype = df_nodes_x.iloc[0,2]
        if nodetype == 'Location':
            # print('It is a Location node')
            bus_loc = ["location", "SL"]
        elif nodetype == 'Bus':
            # print('It is a Bus node')
            bus_loc = ["bus", "B"]

    # TODO: figure out the reserve zone for each node, for SPP there are 5 reserve zones and there should be a correspondance with the nodes
    ResZone = 1
    ############################################################################################

    # Read only the DA market

    ndaysmonth = calendar.monthrange(year, month)
    ndaysmonth = int(ndaysmonth[1])

    for ix in range(ndaysmonth):
        day_x = ix+1

        fnameLMP_DA = "DA-LMP-{0:s}-{1:d}{2:02d}{3:02d}0100.csv".format(bus_loc[1], year, month, day_x)
        fnameMCP_DA = "DA-MCP-{0:d}{1:02d}{2:02d}0100.csv".format(year, month, day_x)

        fname_path_LMP_DA = os.path.join(fpath, 'LMP', 'DAM', bus_loc[0], str(year), str(month).zfill(2),fnameLMP_DA)
        fname_path_MCP_DA = os.path.join(fpath, 'MCP', 'DAM', str(year), str(month).zfill(2),fnameMCP_DA)

        if typedat == "lmp" or typedat == "both":
            # DA-LMP-B-201707010100.csv
            # DA-LMP-SL-201707010100.csv
            try:
                df_daLMP = pd.read_csv(fname_path_LMP_DA, index_col=False)
            except FileNotFoundError:
                daLMP = np.empty([0])
                logging.warning('read_spp_data: LMP file missing, returning empty array.')
                break

            daLMP_node_x = df_daLMP.loc[df_daLMP['Pnode'] == node, ['LMP']]
            daLMP_node_x.drop_duplicates(inplace=True)
            daLMP = np.append(daLMP, daLMP_node_x.values)
            if daLMP_node_x.empty:
                return np.empty([0]), np.empty([0]), np.empty([0])

        if typedat == "mcp" or typedat == "both":
            # DA-MCP-201707010100.csv
            try:
                df_daMCP = pd.read_csv(fname_path_MCP_DA, index_col=False)
            except FileNotFoundError:
                daMCPRU = np.empty([0])
                daMCPRD = np.empty([0])
                logging.warning('read_spp_data: MCP file missing, returning empty arrays.')
                break

            # print('Warning -reserve zone not figured out!!!')
            daMCPRU_node_x = df_daMCP.loc[df_daMCP['Reserve Zone'] == str(ResZone), ['RegUP']]
            daMCPRD_node_x = df_daMCP.loc[df_daMCP['Reserve Zone'] == str(ResZone), ['RegDN']]

            daMCPRU = np.append(daMCPRU, daMCPRU_node_x.values)
            daMCPRD = np.append(daMCPRD, daMCPRD_node_x.values)

    return daLMP, daMCPRU, daMCPRD


def read_spp_data_old(fpath, year, month, node, typedat="both"):
    """"
    Reads the historical LMP, regulation capacity, and regulation service (mileage) prices for the year 'year',
    the month 'month' and for the node 'nodeid'. Returns NumPy ndarrays for those three prices.
    :param fpath: A string containing the path to the relevant historical ancillary services data file.
    :param year: An int corresponding to the year of interest
    :param month: An int corresponding to the month of interest (1: Jan., 2: Feb., etc.)
    :param node: A string with the name of the node in SPP
    :param typedat: xxxxxx xxxxxxx
    :param RT_DAM: xxxxxxxxx xxxx
    :return: daLMP, daMCPRU, daMCPRD: NumPy ndarrays containing hourly LMP as well as regulation up and down prices for the SPP market
    """

    ############################################################
    path_nodes_file = '../../es_gui/apps/data_manager/_static/'
    pathf_nodeszones = os.path.join(fpath, path_nodes_file, 'nodes_spp.csv')

    ############################################################
    ############################################################

    pathfile_locations = os.path.join(fpath, 'SPP_locations_list.csv')
    df_locations = pd.read_csv(pathfile_locations, index_col=False)

    df_locations_x = df_locations.loc[df_locations['Pnode']== node, :]

    if not df_locations_x.empty:
        print('It is a Location node')
        bus_loc = ["location", "SL"]
        pnode_x = df_locations_x.iloc[0,0]
    else:
        # pathfile_buses = folderfile + 'SPP_buses_list.csv'
        pathfile_buses = os.path.join(fpath, 'SPP_buses_list.csv')
        df_buses = pd.read_csv(pathfile_buses, index_col=False)
        df_buses_x = df_buses.loc[df_buses['Pnode'] == node, :]
        if not df_buses_x.empty:
            print('It is a Bus node')
            bus_loc = ["bus", "B"]
            pnode_x = df_buses_x.iloc[0,0]
        else:
            raise ValueError('Unknown Node!!!')

    # TODO: figure out the reserve zone for each node, for SPP there are 5 reserve zones and there should be a correspondance with the nodes
    # Reserve Zone:
    ResZone = 1


    # Read only the DA market

    ndaysmonth = calendar.monthrange(year, month)
    ndaysmonth = int(ndaysmonth[1])

    daLMP = np.empty([0])
    daMCPRU = np.empty([0])
    daMCPRD = np.empty([0])

    for ix in range(ndaysmonth):
        day_x = ix+1

        fnameLMP_DA = "DA-LMP-{0:s}-{1:d}{2:02d}{3:02d}0100.csv".format(bus_loc[1], year, month, day_x)
        fnameMCP_DA = "DA-MCP-{0:d}{1:02d}{2:02d}0100.csv".format(year, month, day_x)

        fname_path_LMP_DA = os.path.join(fpath, 'LMP', 'DAM', bus_loc[0], str(year), str(month).zfill(2),fnameLMP_DA)
        fname_path_MCP_DA = os.path.join(fpath, 'MCP', 'DAM', str(year), str(month).zfill(2),fnameMCP_DA)

        if typedat == "lmp" or typedat == "both":
            # DA-LMP-B-201707010100.csv
            # DA-LMP-SL-201707010100.csv
            try:
                df_daLMP = pd.read_csv(fname_path_LMP_DA, index_col=False)
            except FileNotFoundError:
                daLMP = np.empty([0])
                logging.warning('read_spp_data: LMP file missing, returning empty array.')
                break

            daLMP_node_x = df_daLMP.loc[df_daLMP['Pnode'] == node, ['LMP']]
            daLMP = np.append(daLMP, daLMP_node_x.values)

        if typedat == "mcp" or typedat == "both":
            # DA-MCP-201707010100.csv
            try:
                df_daMCP = pd.read_csv(fname_path_MCP_DA, index_col=False)
            except FileNotFoundError:
                daMCPRU = np.empty([0])
                daMCPRD = np.empty([0])
                logging.warning('read_spp_data: MCP file missing, returning empty arrays.')
                break

            print('Warning -reserve zone not figured out!!!')
            daMCPRU_node_x = df_daMCP.loc[df_daMCP['Reserve Zone'] == ResZone, ['RegUP']]
            daMCPRD_node_x = df_daMCP.loc[df_daMCP['Reserve Zone'] == ResZone, ['RegDN']]

            daMCPRU = np.append(daMCPRU, daMCPRU_node_x.values)
            daMCPRD = np.append(daMCPRD, daMCPRD_node_x.values)

    return daLMP, daMCPRU, daMCPRD
#######################################################################################################################

#######################################################################################################################
# CAISO
#######################################################################################################################

def read_caiso_data(fpath, year, month, nodeid):
    """"
    Reads the historical LMP, regulation up/down and regulation mileage up/down for the year 'year',
    the month 'month' and for the node 'nodeid'. Returns NumPy ndarrays for those three prices.

    :param fpath: The path to the root of the PJM data directory
    :type fpath: str
    :param year: Year of data to read
    :type year: int or str
    :param month: Month of data to read
    :type month: int or str
    :param nodeid: ID of the node to read
    :type nodeid: str
    :return: daLMP, RegCCP, RegPCP: Hourly LMP and regulation capacity/performance clearing price values.
    :rtype: NumPy ndarrays
    """

    """"
    For CAISO certain prices have:
    _CAISO
    _CAISO_EXP
    _NP26
    _NP26_EXP
    _SP26
    _SP26_EXP

    we only use the _CAISO_EXP ones (for mileage prices they are the only ones)

    AS_CAISO_EXP_RD_CLR_PRC
    AS_CAISO_EXP_RU_CLR_PRC

    AS_CAISO_EXP_RMD_CLR_PRC
    AS_CAISO_EXP_RMU_CLR_PRC
    """

    daLMP = np.empty([0])
    daREGU = np.empty([0])
    daREGD = np.empty([0])
    daRMU = np.empty([0])
    daRMD = np.empty([0])
    RMU_MM = np.empty([0])
    RMD_MM = np.empty([0])
    RMU_PACC = np.empty([0])
    RMD_PACC = np.empty([0])

    if isinstance(month, str):
        month = int(month)

    if isinstance(year, str):
        year = int(year)

    # Names examples:
    # 201601_dalmp_LAKESID2_7_UNITS-APND.csv
    # 201601_asp.csv
    # 201601_regm.csv
    fnameLMP = "{0:d}{1:02d}_dalmp_{2:s}.csv".format(year, month, nodeid)
    fnameASP = "{0:d}{1:02d}_asp.csv".format(year, month)
    fnameMILEAGE = "{0:d}{1:02d}_regm.csv".format(year, month)

    fname_path_LMP = os.path.join(fpath, 'LMP', str(nodeid), str(year), fnameLMP)
    fname_path_ASP = os.path.join(fpath, 'ASP', str(year), fnameASP)
    fname_path_MILEAGE = os.path.join(fpath, 'MILEAGE', str(year), fnameMILEAGE)

    try:
        dfLMP = pd.read_csv(fname_path_LMP, index_col=False)
        daLMP = dfLMP['LMP'].values
    except FileNotFoundError:
        logging.warning \
            (
                'read_caiso_data: No LMP data matching input parameters found, returning empty array. (got {fname}, {year}, {month}, {nodeid})'.format
                (fname=fnameLMP, year=year, month=month, nodeid=nodeid))

    try:
        dfASP = pd.read_csv(fname_path_ASP, index_col=False)
        daREGU = dfASP['AS_CAISO_EXP_RU_CLR_PRC'].values
        daREGD = dfASP['AS_CAISO_EXP_RD_CLR_PRC'].values
        daRMU = dfASP['AS_CAISO_EXP_RMU_CLR_PRC'].values
        daRMD = dfASP['AS_CAISO_EXP_RMD_CLR_PRC'].values

    except FileNotFoundError:
        logging.warning \
            (
                'read_caiso_data: No ASP data matching input parameters found, returning empty array. (got {fname}, {year}, {month})'.format
                (fname=fnameASP, year=year, month=month))

    try:
        dfMIL_ACC = pd.read_csv(fname_path_MILEAGE, index_col=False)
        RMU_MM = dfMIL_ACC['RMU_SYS_MIL_MUL'].values
        RMD_MM = dfMIL_ACC['RMD_SYS_MIL_MUL'].values
        RMU_PACC = dfMIL_ACC['RMU_SYS_PERF_ACC'].values
        RMD_PACC = dfMIL_ACC['RMD_SYS_PERF_ACC'].values

    except FileNotFoundError:
        logging.warning \
            (
                'read_caiso_data: No MILEAGE and PERFORMANCE ACCURACY data matching input parameters found, returning empty array. (got {fname}, {year}, {month})'.format
                (fname=fnameMILEAGE, year=year, month=month))

    return daLMP, daREGU, daREGD, daRMU, daRMD, RMU_MM, RMD_MM, RMU_PACC, RMD_PACC
    # TODO: understand the units of the MILEAGE MULTIPLIERS

#######################################################################################################################

#######################################################################################################################
#######################################################################################################################









if __name__ == '__main__':
    from es_gui.tools.valuation.valuation_optimizer import ValuationOptimizer

    fpath = os.path.join('data', 'PJM')
    year = 2015
    month = 1
    nodeid = '1'

    for month in [12]:
        daLMP, mr, rega, regd, RegCCP, RegPCP = read_pjm_data(fpath,year,month,nodeid)

        op = ValuationOptimizer(market_type='pjm_pfp')

        op.price_electricity = daLMP
        op.mileage_ratio = mr
        op.mileage_slow = rega
        op.mileage_fast = regd
        op.price_reg_capacity = RegCCP
        op.price_reg_performance = RegPCP

        handler_requests = {}
        handler_requests['iso'] = 'PJM'
        handler_requests['market type'] = 'pjm_pfp'
        handler_requests['months'] = [(month, year),]
        handler_requests['node id'] = nodeid

        results, gross_revenue = op.run()
        print(gross_revenue)



from __future__ import absolute_import

import os

from kivy.uix.screenmanager import Screen

from es_gui.resources.widgets.common import NavigationButton
from es_gui.proving_grounds.help_carousel import HelpCarouselModalView


class TechSelectionHomeScreen(Screen):
    """The home screen for performing energy storage technology selection analysis."""

    def __init__(self, **kwargs):
        super(TechSelectionHomeScreen, self).__init__(**kwargs)

    def on_enter(self):
        ab = self.manager.nav_bar
        ab.reset_nav_bar()
        ab.set_title('Energy Storage Technology Selection Application')

        help_button = NavigationButton(text='help', on_release=self.open_help_carousel)
        ab.action_view.add_widget(help_button)

    def open_help_carousel(self, *args):
        help_carousel_view = HelpCarouselModalView()
        help_carousel_view.title.text = 'QuESt Technology Selection'

        slide_01_text = ('QuESt Technology Selection is an application with tools for identifying feasible storage '
                         'technlogies for a given project.')
        slide_02_text = ('The technology selection wizard identifies which energy storage technologies satisfy the minimum '
                         'application requirements (such as discharge duration and response time) and rank them. The user '
                         'is required to indicate the grid location where the storage system will be deployed, which set '
                         'default values for the other types of inputs.')
        slide_03_text = ('Upon completion of the wizard, you will be taken to the first results screen, which indicates '
                         'whether each energy storage technology is a feasible option for the desired project.')
        slide_04_text = ('The next results screen displays a rank of the feasible technologies based on four factors. The '
                         'user can modify the weights assigned to each one of these factors, as well as the desired target '
                         'cost (the latter affects the Cost Score). These adjustments will be reflected in the new Total '
                         'Score for each technology.')
        slide_05_text = 'Finally, the user can export any of the previous results to .png or .csv files.'

        slides = [
            (os.path.join('es_gui', 'resources', 'help_views', 'tech_selection', '01.png'), slide_01_text),
            (os.path.join('es_gui', 'resources', 'help_views', 'tech_selection', '03.png'), slide_02_text),
            (os.path.join('es_gui', 'resources', 'help_views', 'tech_selection', '05.png'), slide_03_text),
            (os.path.join('es_gui', 'resources', 'help_views', 'tech_selection', '06.png'), slide_04_text),
            (os.path.join('es_gui', 'resources', 'help_views', 'tech_selection', '07.png'), slide_05_text),
        ]

        help_carousel_view.add_slides(slides)
        help_carousel_view.open()


"""Market data processing tools."""
# from __future__ import absolute_import
# from xlrd.biffh import XLRDError

import pandas as pd
import numpy as np
import os
import calendar
import logging


def read_spp_data(fpath, year, month, node, typedat="both"):
    """Read the historical LMP, regulation capacity, and regulation service (mileage) prices and return NumPy ndarrays for those three prices.

    :param fpath: A string containing the path to the relevant historical ancillary services data file.
    :param year: An int corresponding to the year of interest
    :param month: An int corresponding to the month of interest (1: Jan., 2: Feb., etc.)
    :param node: A string with the name of the node in SPP
    :param typedat: xxxxxx xxxxxxx
    :param RT_DAM: xxxxxxxxx xxxx
    :return: daLMP, daMCPRU, daMCPRD: NumPy ndarrays containing hourly LMP as well as regulation up and down prices for the SPP market
    """
    daLMP = np.empty([0])
    daMCPRU = np.empty([0])
    daMCPRD = np.empty([0])

    if isinstance(month, str):
        month = int(month)

    if isinstance(year, str):
        year = int(year)

    ############################################################################################
    # TODO: path_nodes_file is a folder to adjust when integrating it to QuESt
    path_nodes_file = "../_static/nodes_spp.csv"
    pathf_nodeszones = os.path.join(fpath, path_nodes_file, "nodes_spp.csv")
    print(pathf_nodeszones)
    df_nodes = pd.read_csv(path_nodes_file, index_col=False, encoding="cp1252")
    df_nodes_x = df_nodes.loc[df_nodes["Node ID"] == node, :]

    if df_nodes_x.empty:
        print("The node does NOT exists in SPP")
        # raise ValueError('Not a valid bus number!!!')
        return daLMP, daMCPRU, daMCPRD
    else:
        nodetype = df_nodes_x.iloc[0, 2]
        if nodetype == "Location":
            # print('It is a Location node')
            bus_loc = ["location", "SL"]
        elif nodetype == "Bus":
            # print('It is a Bus node')
            bus_loc = ["bus", "B"]

    # TODO: figure out the reserve zone for each node, for SPP there are 5 reserve zones and there should be a correspondance with the nodes
    ResZone = 1
    ############################################################################################

    # Read only the DA market

    ndaysmonth = calendar.monthrange(year, month)
    ndaysmonth = int(ndaysmonth[1])

    for ix in range(ndaysmonth):
        day_x = ix + 1

        fnameLMP_DA = "DA-LMP-{0:s}-{1:d}{2:02d}{3:02d}0100.csv".format(
            bus_loc[1], year, month, day_x
        )
        fnameMCP_DA = "DA-MCP-{0:d}{1:02d}{2:02d}0100.csv".format(year, month, day_x)

        fname_path_LMP_DA = os.path.join(
            fpath, "LMP", "DAM", bus_loc[0], str(year), str(month).zfill(2), fnameLMP_DA
        )
        fname_path_MCP_DA = os.path.join(
            fpath, "MCP", "DAM", str(year), str(month).zfill(2), fnameMCP_DA
        )

        if typedat == "lmp" or typedat == "both":
            # DA-LMP-B-201707010100.csv
            # DA-LMP-SL-201707010100.csv
            try:
                df_daLMP = pd.read_csv(fname_path_LMP_DA, index_col=False)
            except FileNotFoundError:
                daLMP = np.empty([0])
                logging.warning(
                    "read_spp_data: LMP file missing, returning empty array."
                )
                break

            daLMP_node_x = df_daLMP.loc[df_daLMP["Pnode"] == node, ["LMP"]]
            daLMP = np.append(daLMP, daLMP_node_x.values)
            if daLMP_node_x.empty:
                return np.empty([0]), np.empty([0]), np.empty([0])

        if typedat == "mcp" or typedat == "both":
            # DA-MCP-201707010100.csv
            try:
                df_daMCP = pd.read_csv(fname_path_MCP_DA, index_col=False)
            except FileNotFoundError:
                daMCPRU = np.empty([0])
                daMCPRD = np.empty([0])
                logging.warning(
                    "read_spp_data: MCP file missing, returning empty arrays."
                )
                break

            # print('Warning -reserve zone not figured out!!!')
            daMCPRU_node_x = df_daMCP.loc[
                df_daMCP["Reserve Zone"] == ResZone, ["RegUP"]
            ]
            daMCPRD_node_x = df_daMCP.loc[
                df_daMCP["Reserve Zone"] == ResZone, ["RegDN"]
            ]

            daMCPRU = np.append(daMCPRU, daMCPRU_node_x.values)
            daMCPRD = np.append(daMCPRD, daMCPRD_node_x.values)

    return daLMP, daMCPRU, daMCPRD


from __future__ import absolute_import

import logging
from datetime import datetime
import calendar
import pyutilib
import numpy as np

from kivy.clock import mainthread

from es_gui.tools.btm.btm_optimizer import BtmOptimizer, BadParameterException, IncompatibleDataException
import es_gui.tools.btm.readutdata as readutdata


class BtmOptimizerHandler:
    """A handler for creating and solving BtmOptimizer instances as requested."""
    dms = None
    solved_ops = []

    def __init__(self, solver_name):
        self._solver_name = solver_name

    @property
    def solver_name(self):
        """The name of the solver for Pyomo to call."""
        return self._solver_name

    @solver_name.setter
    def solver_name(self, value):
        self._solver_name = value

    def process_requests(self, op_handler_requests, *args):
        """Generates and solves BtmOptimizer models based on the given requests."""
        dms = self.dms

        rate_structure = op_handler_requests['rate_structure']
        load_profile_path = op_handler_requests['load_profile']
        pv_profile_path = op_handler_requests['pv_profile']
        param_set = op_handler_requests['params']

        solved_requests = []

        handler_status = set()

        year = 2019

        weekday_energy_schedule = rate_structure['energy rate structure']['weekday schedule']
        weekend_energy_schedule = rate_structure['energy rate structure']['weekend schedule']
        weekday_demand_schedule = rate_structure['demand rate structure']['weekday schedule']
        weekend_demand_schedule = rate_structure['demand rate structure']['weekend schedule']

        nem_type = 2 if rate_structure['net metering']['type'] else 1
        nem_rate = None if rate_structure['net metering']['type'] else rate_structure['net metering']['energy sell price']

        rate_df = readutdata.input_df(year, weekday_energy_schedule, weekend_energy_schedule, weekday_demand_schedule, weekend_demand_schedule)

        for ix, month in enumerate(calendar.month_abbr[1:], start=1):
            param_set_iterator = iter(param_set)
            continue_param_loop = True

            while continue_param_loop:
                try:
                    params = next(param_set_iterator)
                except StopIteration:
                    break

                op = BtmOptimizer()

                # Get data.
                # TODO: Move to a DMS. Should the omission of PV profile data be handled by the BtmOptimizer?
                load_profile = self.dms.get_load_profile_data(load_profile_path['path'], ix)

                try:
                    pv_profile = self.dms.get_pv_profile_data(pv_profile_path['path'], ix)
                except KeyError:
                    pv_profile = np.zeros(len(load_profile))

                # Build op inputs.
                rate_df_month = rate_df.loc[rate_df['month'] == ix]

                # Populate op.
                op.tou_energy_schedule = rate_df_month['tou_energy_schedule'].values
                op.tou_demand_schedule = rate_df_month['tou_demand_schedule'].values

                op.tou_energy_rate = [x[1] for x in rate_structure['energy rate structure']['energy rates'].items()]
                op.tou_demand_rate = [x[1] for x in rate_structure['demand rate structure']['time of use rates'].items()]
                op.flat_demand_rate = rate_structure['demand rate structure']['flat rates'][month]

                op.nem_type = nem_type
                op.nem_rate = nem_rate

                op.load_profile = load_profile
                op.pv_profile = pv_profile
                op.rate_structure_metadata = rate_structure
                op.load_profile_metadata = load_profile_path
                op.pv_profile_metadata = pv_profile_path

                if params:
                    op.set_model_parameters(**params)
                else:
                    continue_param_loop = False
                
                try:
                    solved_op = self._solve_model(op)
                except pyutilib.common._exceptions.ApplicationError as e:
                    logging.error('Op Handler: {error}'.format(error=e))

                    if 'No executable found' in e.args[0]:
                        # Could not locate solver executable
                        handler_status.add('* The executable for the selected solver could not be found; please check your installation.')
                    else:
                        handler_status.add('* ({0} {1}) {2}. The problem may be infeasible.'.format(month, year, e.args[0]))
                except IncompatibleDataException as e:
                    # Data exception raised by BtmOptimizer
                    logging.error(e)
                    handler_status.add('* ({0} {1}) The time series data has mismatched sizes.'.format(month, year))
                except AssertionError as e:
                    # An optimal solution could not be found as reported by the solver
                    logging.error('Op Handler: {error}'.format(error=e))
                    handler_status.add('* ({0} {1}) An optimal solution could not be found; the problem may be infeasible.'.format(month, year))
                else:
                    solved_op = self._save_to_solved_ops(solved_op, month, params)
                    solved_requests.append(solved_op)

        logging.info('Op Handler: Finished processing requested jobs.')
        return solved_requests, handler_status

    def _solve_model(self, op):
        op.solver = self.solver_name
        op.run()

        return op

    @staticmethod
    def _save_to_solved_ops(op, month, param_set):
        # time_finished = datetime.now().strftime('%A, %B %d, %Y %H:%M:%S')
        time_finished = datetime.now().strftime('%b %d, %Y %H:%M:%S')

        name_components = [time_finished, month,]

        # Check for rate structure name.
        rate_structure_name = 'Rate: {0}'.format(op.rate_structure_metadata.get('name', ''))
        name_components.append(rate_structure_name)

        # Check for PV profile.
        if any(op.pv_profile != 0):
            if op.pv_profile_metadata:
                pv_profile_name = 'PV: {0}'.format(op.pv_profile_metadata.get('name', 'Not specified'))
            else:
                pv_profile_name = 'PV: Not specified'
            
            name_components.append(pv_profile_name)
        else:
            name_components.append('PV: None')
        
        # Load profile name.
        load_profile_name = 'Load: {0}'.format(op.load_profile_metadata.get('name', ''))
        name_components.append(load_profile_name)

        name = ' | '.join(name_components)

        results_dict = {}

        results_dict['name'] = name
        results_dict['optimizer'] = op
        results_dict['month'] = month
        if param_set:
            results_dict['params'] = param_set
        results_dict['time'] = time_finished
        results_dict['label'] = ' '.join([month, repr(param_set)])

        BtmOptimizerHandler.solved_ops.append(results_dict)

        return (name, op)
    
    def get_solved_ops(self):
        """Returns the list of solved Optimizer objects in reverse chronological order."""
        return_list = reversed(self.solved_ops)

        return return_list
    


"""Market data processing tools."""
# from __future__ import absolute_import
# from xlrd.biffh import XLRDError

import pandas as pd
import numpy as np
import os
import logging


def read_nodeid(fname, iso):
    """Read node id from an excel file and return as a list."""
    from xlrd import open_workbook

    wb = open_workbook(filename=fname)
    ws = wb.sheet_by_name(iso)
    nodeid = []
    for i in range(1, ws.nrows - 1):
        cell = ws.cell(i, 0)
        text = str(cell.value)
        text = text.replace(".0", "")
        nodeid += [text]
    return nodeid

def read_isone_data(fpath, year, month, nodeid):
    """Read ISONE historical LMP, regulation capacity, and regulation service (mileage) prices and return NumPy ndarrays for those three prices.

    :param fpath: A string containing the path to the relevant historical ancillary services data file.
    :param year: An int corresponding to the year of interest
    :param month: An int corresponding to the month of interest (1: Jan., 2: Feb., etc.)
    :return: daLMP, RegCCP, RegPCP: NumPy ndarrays containing hourly LMP as well as regulation capacity/performance clearing price values.
    """
    if isinstance(month, str):
        month = int(month)

    if isinstance(year, str):
        year = int(year)

    if isinstance(nodeid, (int, float, complex)):
        nodeid = str(nodeid)

    if year < 2018:
        if year == 2017 and month == 12:
            fnameLMP = "{0:d}{1:02d}_fmlmp_{2:s}.csv".format(year, month, nodeid)
            fnameRCP = "{0:d}{1:02d}_fmrcp.csv".format(year, month)
        else:
            fnameLMP = "{0:d}{1:02d}_dalmp_{2:s}.csv".format(year, month, nodeid)
            fnameRCP = "{0:d}{1:02d}_rcp.csv".format(year, month)
    else:
        fnameLMP = "{0:d}{1:02d}_fmlmp_{2:s}.csv".format(year, month, nodeid)
        fnameRCP = "{0:d}{1:02d}_fmrcp.csv".format(year, month)

    fname_path_LMP = os.path.join(fpath, "LMP", str(nodeid), str(year), fnameLMP)
    fname_path_RCP = os.path.join(fpath, "RCP", str(year), fnameRCP)
    fname_path_MILEAGE = os.path.join(fpath, "MileageFile.xlsx")

    daLMP = np.empty([0])
    RegCCP = np.empty([0])
    RegPCP = np.empty([0])
    miMULT = np.empty([0])

    try:
        dfLMP = pd.read_csv(fname_path_LMP, index_col=False)
        daLMP = dfLMP["LmpTotal"].values
    except FileNotFoundError:
        logging.warning(
            "read_isone_data: No LMP data matching input parameters found, returning empty array. (got {fname}, {year}, {month}, {nodeid})".format(
                fname=fnameLMP, year=year, month=month, nodeid=nodeid
            )
        )

    try:
        if year > 2014:
            if year == 2015 and month < 4:
                dfRCP = pd.read_csv(fname_path_RCP, index_col=False)
                RegCCP = dfRCP["RegClearingPrice"].values
                RegPCP = []
            else:
                dfRCP = pd.read_csv(fname_path_RCP, index_col=False)
                RegCCP = dfRCP["RegCapacityClearingPrice"].values
                RegPCP = dfRCP["RegServiceClearingPrice"].values

                dataF_mileage_file = pd.read_excel(
                    fname_path_MILEAGE,
                    sheet_name="Energy Neutral Trinary",
                    usecols=["Fleet ATRR dispatch [MW]"],
                )
                dataF_mileage_file = dataF_mileage_file.append(
                    pd.DataFrame([-10] * 15, columns=["Fleet ATRR dispatch [MW]"]),
                    ignore_index=True,
                )  # Change number of data points to 24 hours; doesn't change mileage

                #   AGC setpoints given every 4 seconds, take the total mileage for each hour; total of one day of mileage
                hours = [i for i in range(len(dataF_mileage_file.index) // 900)]

                mileage_day = []
                for hour in hours:
                    dataF_mileage_hour = dataF_mileage_file[
                        900 * hour : 900 * (hour + 1)
                    ]  # Every 900 values represents an hour (900*4 = 3600)
                    mileage_hour = 0
                    for i in range(len(dataF_mileage_hour.index)):
                        if i == len(dataF_mileage_hour.index) - 1:
                            break

                        if (
                            not dataF_mileage_hour.iloc[i, 0]
                            == dataF_mileage_hour.iloc[i + 1, 0]
                        ):
                            mileage_hour += (
                                abs(
                                    dataF_mileage_hour.iloc[i, 0]
                                    - dataF_mileage_hour.iloc[i + 1, 0]
                                )
                                / 10
                            )

                    mileage_day.append(mileage_hour)

                dataF_mileage_day = pd.DataFrame(
                    mileage_day, columns=["Trinary Mileage"]
                )

                #   have one days worth of data, need one months worth
                days = len(daLMP) // 24
                mileage_mult = pd.DataFrame(columns=["Trinary Mileage"])
                for day in range(days):
                    mileage_mult = mileage_mult.append(
                        dataF_mileage_day, ignore_index=True
                    )
                #   if the len are offset, make them match
                if not len(daLMP) == len(mileage_mult):
                    diff = len(daLMP) - len(mileage_mult)

                    for i in range(diff):
                        mileage_mult = mileage_mult.append(
                            dataF_mileage_day.iloc[i], ignore_index=True
                        )

                miMULT = mileage_mult["Trinary Mileage"].values
        else:
            dfRCP = pd.read_csv(fname_path_RCP, index_col=False)
            RegCCP = dfRCP["RegClearingPrice"].values
            RegPCP = []

    except FileNotFoundError:
        logging.warning(
            "read_isone_data: No ASP data matching input parameters found, returning empty array. (got {fname}, {year}, {month})".format(
                fname=fnameRCP, year=year, month=month
            )
        )

    return daLMP, RegCCP, RegPCP, miMULT


"""Market data processing tools."""
# from __future__ import absolute_import
# from xlrd.biffh import XLRDError

import pandas as pd
import numpy as np
import os
import logging


def read_nodeid(fname, iso):
    """Read node id from an excel file and return as a list."""
    from xlrd import open_workbook

    wb = open_workbook(filename=fname)
    ws = wb.sheet_by_name(iso)
    nodeid = []
    for i in range(1, ws.nrows - 1):
        cell = ws.cell(i, 0)
        text = str(cell.value)
        text = text.replace(".0", "")
        nodeid += [text]
    return nodeid


def read_pjm_data(fpath, year, month, nodeid):
    """Read the historical LMP, regulation capacity, and regulation service (mileage) prices in PJM and returns NumPy ndarrays for those three prices.

    :param fpath: The path to the root of the PJM data directory
    :type fpath: str
    :param year: Year of data to read
    :type year: int or str
    :param month: Month of data to read
    :type month: int or str
    :return: daLMP, RegCCP, RegPCP: Hourly LMP and regulation capacity/performance clearing price values.
    :rtype: NumPy ndarrays
    """
    daLMP = np.array([])
    RegCCP = np.array([])
    RegPCP = np.array([])
    rega = np.array([])
    regd = np.array([])
    mr = np.array([])

    if isinstance(month, str):
        month = int(month)

    if isinstance(year, str):
        year = int(year)

    if isinstance(nodeid, str):
        nodeid = int(nodeid)

    fnameLMP = "{0:d}{1:02d}_dalmp_{2:d}.csv".format(year, month, nodeid)
    fnameREG = "{0:d}{1:02d}_regp.csv".format(year, month)
    fnameMILEAGE = "{0:d}{1:02d}_regm.csv".format(year, month)

    fname_path_LMP = os.path.join(fpath, "LMP", str(nodeid), str(year), fnameLMP)
    fname_path_REG = os.path.join(fpath, "REG", str(year), fnameREG)
    fname_path_MILEAGE = os.path.join(fpath, "MILEAGE", str(year), fnameMILEAGE)

    try:
        dfLMP = pd.read_csv(fname_path_LMP, index_col=False)
        daLMP = dfLMP["total_lmp_da"].values
    except FileNotFoundError:
        logging.warning(
            "read_pjm_data: No LMP data matching input parameters found, returning empty array. (got {fname}, {year}, {month}, {nodeid})".format(
                fname=fnameLMP, year=year, month=month, nodeid=nodeid
            )
        )

    try:
        dfREG = pd.read_csv(fname_path_REG, index_col=False)
        RegCCP = dfREG["rmccp"].values
        RegPCP = dfREG["rmpcp"].values
    except FileNotFoundError:
        logging.warning(
            "read_pjm_data: No REG data matching input parameters found, returning empty array. (got {fname}, {year}, {month})".format(
                fname=fnameREG, year=year, month=month
            )
        )

    try:
        dfMILEAGE = pd.read_csv(fname_path_MILEAGE, index_col=False)
        dfMILEAGE["MILEAGE RATIO"] = dfMILEAGE["regd_hourly"] / dfMILEAGE["rega_hourly"]

        # TODO: Handling NaNs/missing data intelligently. The current method just fills forward.
        rega = (
            dfMILEAGE["rega_hourly"]
            .replace([np.inf, -np.inf], np.nan)
            .fillna(method="ffill")
            .values
        )
        regd = (
            dfMILEAGE["regd_hourly"]
            .replace([np.inf, -np.inf], np.nan)
            .fillna(method="ffill")
            .values
        )
        mr = (
            dfMILEAGE["MILEAGE RATIO"]
            .replace([np.inf, -np.inf], np.nan)
            .fillna(method="ffill")
            .values
        )
    except FileNotFoundError:
        logging.warning(
            "read_pjm_data: No MILEAGE data matching input parameters found, returning empty array. (got {fname}, {year}, {month})".format(
                fname=fnameMILEAGE, year=year, month=month
            )
        )

    return daLMP, mr, rega, regd, RegCCP, RegPCP


def read_pjm_da_lmp(fname, node_name):
    """Read PJM day-ahead LMP file and return the NumPy ndarray corresponding to the hourly LMP at node_name.

    :param fname: A string containing the path to the relevant day-ahead LMP file.
    :param node_name: A string containing the name of the pricing node of interest.
    :return: LMP: A NumPy ndarray containing the hourly LMP at node-name.
    """
    # read in the .csv file
    df = pd.read_csv(fname, low_memory=False)

    # filter rows by node_name
    col2 = df.axes[1][2]
    pnode_ix = df.index[df[col2] == node_name]
    df1 = df.iloc[pnode_ix, :]

    # filter Total LMP columns
    df2 = df1[df1.axes[1][7:79:3]]

    # convert to NumPy ndarray, ravel, and remove NaNs
    LMP = np.ravel(df2.astype("float").values)
    LMP = LMP[~np.isnan(LMP)]

    return LMP


def read_pjm_reg_signal(fname):
    """
    Read PJM regulation signal file at and return the NumPy ndarray corresponding to the hourly integrated signal.

    :param fname: A string containing the path to the relevant regulation signal file.
    :return: RU, RD: NumPy ndarrays containing the hourly integrated regulation up/down signals.
    """
    # read in the Excel file
    df = pd.read_excel(fname, skip_footer=1)

    # create DateTime indexing to facilitate resampling
    dt_ix = pd.date_range("2017-11-01", periods=30 * 60 * 24, freq="2S")
    df.index = dt_ix

    # define function for performing hourly integration
    def _hourly_integration(array_like):
        # ZOH integration
        dt = 2.0 / (60 * 60)
        return np.sum(array_like) * dt

    # use resample to apply hourly integration
    df1 = df.resample(rule="H", closed="left").apply(_hourly_integration)

    # convert DataFrame to NumPy ndarray and ravel
    REG = np.ravel(df1.astype("float").values, "F")

    # assign reg up/down values appropriately based on sign of regulation signal
    RU = REG * (REG >= 0)
    RD = REG * (REG < 0)

    return RU, RD


def read_pjm_mileage(fname, month):
    """
    Read PJM historical regulation market data file at fname and returns the NumPy ndarrays for mileage data.

    :param fname: A string containing the path to the relevant historic regulation market data file.
    :param month: An int corresponding to the month of interest (1: Jan., 2: Feb., etc.)
    :return: Mileage_Ratio, RegA_milage, RegD_mileage: NumPy ndarrays containing computed mileage ratio and RegA/RegD hourly mileage signals.
    """
    # read in the Excel file and parse the relevant worksheet
    wkbk = pd.ExcelFile(fname)
    df = wkbk.parse(month)

    # replace "UNAPPROVED" entries with previous filled value
    df.fillna(method="ffill", inplace=True)

    # compute hourly mileage ratio
    # TODO: what do infinite mileage ratio? (REGA = 0)
    df["MILEAGE RATIO"] = df["REGD_HOURLY"] / df["REGA_HOURLY"]

    RegA_mileage = df["REGA_HOURLY"].values
    RegD_mileage = df["REGD_HOURLY"].values
    Mileage_Ratio = df["MILEAGE RATIO"].values

    return Mileage_Ratio, RegA_mileage, RegD_mileage


def read_pjm_reg_price(fname, month):
    """
    Read PJM historical ancillary services data file and return the NumPy ndarrays for regulation clearing prices.

    :param fname: A string containing the path to the relevant historical ancillary services data file.
    :param month: An int corresponding to the month of interest (1: Jan., 2: Feb., etc.)
    :return: RegCCP, RegPCP: NumPy ndarrays containing hourly regulation capacity/performance clearing price values.
    """
    # read in the Excel file and parse relevant worksheet
    wkbk = pd.ExcelFile(fname)
    df = wkbk.parse(month)

    # parse the relevant service
    df1 = df[df["SERVICE"] == "REG"]

    # DLS hour gives NaN
    df1.fillna(method="ffill", inplace=True)

    RegCCP = df1["REG_CCP"].values
    RegPCP = df1["REG_PCP"].values

    return RegCCP, RegPCP


# import pandas as pd
from es_gui.apps.tech_selection import fAux


def isFeasibleTech_Location(techDB, selected_location):
    """
    Identify which ES techs are feasible for the current project based on the location where the ESS will be
    connected to the electric grid.

    Parameters
    ----------
    techDB: pandas dataframe
        Database containing detailed information for multiple ES techs
    selected_location: string
        Location where the ESS will be connected to the electric grid; valid values are 'Transmission/Central',
        'Distribution', 'BTM: Commercial/industrial', and 'BTM: residential'

    Returns
    -------
    List with a subset of ES techs that satisfy the grid location restriction
    """

    # Return the subset of ES techs that satisfy the grid location restriction
    return techDB.loc[fAux.pdSeriesIdxWhereTrue(techDB[f'Score for {selected_location}'] > 0)]['Storage technology (short name)'].values


def isFeasibleTech_Duration(techDB, selected_duration, remaining_feasible_techs=None):
    """
    Identify which ES techs are feasible for the current project based on the minimum discharge duration required
    for the selected application.

    Parameters
    ----------
    techDB: pandas dataframe
        Database containing detailed information for multiple ES techs
    selected_duration: float
        Minimum discharge duration required for the selected application
    remaining_feasible_techs: list, optional
        List of ES techs to be evaluated for feasibility; the default is None (use all ES techs in the database)

    Returns
    -------
    List with a subset of ES techs that satisfy the minimum discharge duration requirement
    """

    # If no subset of ES techs is provided, then evaluate all ES techs in the database
    if remaining_feasible_techs is None:
        remaining_feasible_techs = techDB.index

    # Return the subset of ES techs that satisfy the minimum discharge duration requirement
    return techDB.loc[fAux.pdSeriesIdxWhereTrue(techDB.loc[remaining_feasible_techs, 'Discharge duration (hours)'] >=
                                     selected_duration)]['Storage technology (short name)'].values


def isFeasibleTech_ResponseTime(techDB, selected_response_time, remaining_feasible_techs=None):
    """
    Identify which ES techs are feasible for the current project based on the minimum response time to full power
    required for the selected application.

    Parameters
    ----------
    techDB: pandas dataframe
        Database containing detailed information for multiple ES techs
    selected_response_time: string
        Minimum response time to full power required for the selected application; valid values are 'ms', 'sec', 'min', and'hrs'
    remaining_feasible_techs: list, optional
        List of ES techs to be evaluated for feasibility; the default is None (use all ES techs in the database)

    Returns
    -------
    List with a subset of ES techs that satisfy the minimum response time requirement
    """

    # If no subset of ES techs is provided, then evaluate all ES techs in the database
    if remaining_feasible_techs is None:
        remaining_feasible_techs = techDB.index

    # Ordering of response time values for comparison (ascending order)
    ordering_response_time = ['ms', 'sec', 'min', 'hrs']

    # Convert tech response times into an integer according to the ordering above (for easier comparison)
    tech_response_time_scale = techDB.loc[remaining_feasible_techs, 'Response time to full power'].\
        apply(lambda x: ordering_response_time.index(x))

    # Return the subset of ES techs that satisfy the minimum response time requirement
    return techDB.loc[fAux.pdSeriesIdxWhereTrue(tech_response_time_scale <= 
        ordering_response_time.index(selected_response_time))]['Storage technology (short name)'].values


def isFeasibleTech_ElectricOutput(techDB, requires_electric_output, remaining_feasible_techs=None):
    """
    Identify which ES techs are feasible for the current project based on the need of an output in electric
    form for the selected application.

    Parameters
    ----------
    techDB: pandas dataframe
        Database containing detailed information for multiple ES techs
    requires_electric_output: string
        Indicator of whether the selected application requires an electric output ('Yes' or 'No')
    remaining_feasible_techs: list, optional
        List of ES techs to be evaluated for feasibility; the default is None (use all ES techs)

    Returns
    -------
    List with a subset of ES techs that satisfy the electric output requirement
    """

    # If no subset of ES techs is provided, then evaluate all ES techs in the database
    if remaining_feasible_techs is None:
        remaining_feasible_techs = techDB.index

    # No electric output is required; thus, no ES tech is removed from the feasible subset
    if requires_electric_output == 'No':
        return techDB['Storage technology (short name)'].values
        # try:
            # return remaining_feasible_techs.tolist()
        # except:
            # return remaining_feasible_techs

    # Electric output is required; return the subset of ES techs that satisfy this requirement
    elif requires_electric_output == 'Yes':
        return techDB.loc[fAux.pdSeriesIdxWhereTrue(techDB.loc[remaining_feasible_techs, 'Output in electric form'] == 'Yes')]['Storage technology (short name)'].values


"""This module is for the HelpCarousel widget.

A HelpCarousel is a modal view. It contains a carousel which hosts a series of slides with accompanying text. The primary purpose of this widget is to provide additional help illustrated with screenshots or other relevant figures without overloading the main user interface with information. The modal view includes previous and next buttons to navigate the slides in addition to a group of radio buttons to indicate progress in the carousel's slide deck. The view does not have a dismiss button but auto_dismiss is enabled; the view can be dismissed by clicking outside of it.

The HelpCarouselModalView is designed to be instantiated then populated using the `add_slides()` class method. This method populates the carousel's slides with pairs of image sources and text.
"""

import logging
import os

import pandas as pd

from kivy.uix.checkbox import CheckBox
from kivy.uix.modalview import ModalView
from kivy.uix.boxlayout import BoxLayout
from kivy.properties import StringProperty, BooleanProperty, NumericProperty


class HelpCarouselSlide(BoxLayout):
    """A slide for the HelpCarousel consisting of a large image (80%) and text (20%) in horizontal orientation.
    """
    pass


class HelpCarouselModalView(ModalView):
    """A ModalView with a series of prompts for importing time series data from a csv file.
    """
    current_slide_index = NumericProperty()

    def add_slides(self, slide_deck):
        """Adds image and text to a new slide in the carousel slide deck.

        Each slide consists of a large image on the left and accompanying text on the right.

        Parameters
        ----------
        slide_deck : list(tuple)
            Content for each slide (source, caption) where the source is the path to the slide image and the caption is the text
        
        Notes
        -----
        The source is relative to the current working directory (alongside main.py).
        """
        self.slide_progress_radio_buttons = []

        for source, caption in slide_deck:
            slide = HelpCarouselSlide()
            slide.img.source = source
            slide.img_caption.text = caption

            self.carousel.add_widget(slide)
            self.slide_progress_radio_buttons.append(SlideProgressRadioButton())
            
        for ix, button in enumerate(self.slide_progress_radio_buttons):
            button.active = ix == 0
            self.slide_progress_bx.add_widget(button)
    
    def change_slide(self, direction):
        """Changes carousel slide in the specified direction.
        """
        getattr(self.carousel, 'load_{0}'.format(direction))()

        if direction == 'previous':
            destination_slide = self.carousel.previous_slide
        else:
            destination_slide = self.carousel.next_slide
        
        if destination_slide is not None:
            self.current_slide_index = self.carousel.slides.index(destination_slide)
    
    def on_current_slide_index(self, instance, value):
        # Changes the active button in the slide progress group to reflect the new slide.
        self.slide_progress_radio_buttons[value].active = True


class SlideProgressRadioButton(CheckBox):
    """Radio button representing progress within the HelpCarousel slide deck.
    """
    pass


from __future__ import absolute_import, print_function

import logging
from functools import partial
import os
import pandas as pd

from kivy.uix.screenmanager import Screen, ScreenManager
from kivy.properties import StringProperty
from kivy.clock import Clock
from kivy.app import App

from es_gui.resources.widgets.common import MyPopup, WarningPopup, InputError, fade_in_animation, RecycleViewRow, ValuationRunCompletePopup


class PerformanceSimScreenManager(ScreenManager):
    """Screen Manager for Performance tool."""

    def __init__(self, **kwargs):
        super(PerformanceSimScreenManager, self).__init__(**kwargs)

        # Add new screens here.
        self.add_widget(PerformanceSimDataScreen(name='data'))
        self.add_widget(PerformanceSimParamScreen(name='params'))
        self.add_widget(PerformanceSimSelectionsScreen(name='selections'))


class PerformanceSimRunScreen(Screen):
    """Main screen for Performance simulations."""

    def on_enter(self):
        """Set navigation bar title, check for energyplus, and create popups for running simulations."""
        ab = self.manager.nav_bar
        ab.set_title('Performance Simulations')

        self.check_eplus()

        self.warning_popup = WarningPopup()
        self.run_popup = PerformanceRunPopup()
        self.completion_popup = PerformanceSimCompletePopup()

    def _generate_requests(self):
        """Get the requested data from each screen."""
        data_screen = self.sm.get_screen('data')
        params_screen = self.sm.get_screen('params')

        hvac, location, profile = data_screen.get_inputs()
        param_settings = params_screen.get_inputs()

        requests = {'hvac': hvac,
                    'location': location,
                    'profile': profile,
                    'params': param_settings
                    }

        return requests

    def run_batch(self):
        """Run the performance simulation."""
        try:
            requests = self._generate_requests()
        except ValueError as e:
            self.warning_popup.popup_text.text = str(e)
            self.warning_popup.open()
        except InputError as e:
            self.warning_popup.popup_text.text = str(e)
            self.warning_popup.open()
        else:
            try:
                self.run_popup.run_open()
                self.manager.handler.process_requests(requests)
            except ModuleNotFoundError as e:
                logging.warning('PerformanceSim : {0}'.format(e))
                self.run_popup.dismiss()
                self.warning_popup.popup_text.text = str(e)
                self.warning_popup.open()
            else:
                self.run_popup.run_dismiss()
                self.completion_popup.view_results_button.bind(on_release=self._go_to_view_results)
                self.completion_popup.title = 'Success!'
                self.completion_popup.complete_open()

    def _go_to_view_results(self, *args):
        """Send user to the results viewer."""
        self.manager.nav_bar.go_to_screen('performance_results_viewer')
        self.completion_popup.complete_dismiss()

    def get_valuation_ops(self):
        """Get the completed QuESt Valuation optimization results."""
        valuation_ops = [op for op in self.manager.get_screen('valuation_home').handler.get_solved_ops()]

        valuationF = pd.DataFrame()
        while valuation_ops:
            root = valuation_ops[0]
            root_ls = [op for op in valuation_ops
                       if (op['iso'] == root['iso'] and op['market type'] == root['market type']
                           and op['node'] == root['node'] and op['year'] == root['year'])]
            rootF = pd.DataFrame({root['time']: root_ls})
            for op in root_ls:
                valuation_ops.pop(valuation_ops.index(op))

            valuationF = pd.concat([valuationF, rootF], axis=1)
        else:
            print('Success!')

        valuationF.columns = ['Valuation ' + col for col in valuationF.columns]
        self.sm.get_screen('data').valuation_ops = valuationF

    def get_btm_ops(self):
        """Get the data from all Behind-the-meter optimizations."""
        btm_ops = [op for op in self.manager.get_screen('btm_home').handler.get_solved_ops()]
        for op in btm_ops:
            string = op['name']
            spl_str = string.split(' | ')
            op['rate'] = spl_str[2]
            op['pv'] = spl_str[3]
            op['load'] = spl_str[4]

        btmF = pd.DataFrame()
        while btm_ops:
            root = btm_ops[0]
            root_ls = [op for op in btm_ops
                       if (op['params'] == root['params'] and op['rate'] == root['rate']
                           and op['pv'] == root['pv'] and op['load'] == root['load'])]
            rootF = pd.DataFrame({root['time']: root_ls})
            for op in root_ls:
                btm_ops.pop(btm_ops.index(op))

            btmF = pd.concat([btmF, rootF], axis=1)
        else:
            print('Success!')

        btmF.columns = ['BTM ' + col for col in btmF.columns]
        self.sm.get_screen('data').btm_ops = btmF

    def check_eplus(self):
        """Ensure EnergyPlus and necessary python modules are installed."""
        try:
            from es_gui.apps.performance.performance_sim_handler import PerformanceSimHandler
            data_manager = App.get_running_app().data_manager
            self.manager.handler = PerformanceSimHandler(os.path.join(data_manager.data_bank_root, 'output'))
        except ModuleNotFoundError as e:
            logging.warning('Performance: {}'.format(e))

            popup = EnergyPlusPopup()
            if str(e) == "No module named 'pyenergyplus'":
                popup.open()
            elif str(e) == 'eppy':
                popup.popup_text.text = "This tool requires the python module eppy. Please pip install eppy on your machine."
                popup.open()
            else:
                popup.popup_text.text = "There is a required python module missing." \
                    + " Please visit your console window to determine which module you need to install."
                popup.open()
        else:
            self.get_valuation_ops()
            self.get_btm_ops()
#            self.get_rates()


class PerformanceSimDataScreen(Screen):
    """Data selection screen for Performance Tool."""

    hvac = StringProperty('')
    location = StringProperty('')
    profile = StringProperty('')
    valuation_ops = None
    btm_ops = None

    def __init__(self, **kwargs):
        super(PerformanceSimDataScreen, self).__init__(**kwargs)

        HVACEntry.host_screen = LocationEntry.host_screen = ProfileEntry.host_screen = self

    def on_enter(self):
        """Reset screen on enter."""
        self._reset_screen()

    def _get_hvac_options(self):
        """Get input files from data bank."""
        try:
            data_manager = App.get_running_app().data_manager
            hvac_options = data_manager.data_bank['idf files'].keys()
        except AttributeError:
            pass
        except KeyError:
            if not os.path.exists(os.path.join(data_manager.data_bank_root, 'idf')):
                popup = WarningPopup()
                popup.popup_text.text = "Looks like there are no EnergyPlus input files! Please visit the help tab to learn more."
                popup.open()
            else:
                data_root = os.path.join(data_manager.data_bank_root, 'idf')
                ls = [os.path.join(data_root, item) for item in os.listdir(data_root)]
                bool_ls = [True if os.path.isfile(item) else False for item in ls]

                if any(bool_ls):
                    bad_dir_popup = WarningPopup()
                    bad_dir_popup.popup_text.text = "There was an error processing the HVAC files." \
                        + " Please place your HVAC files in their own folder in the IDF directory."
                    bad_dir_popup.open()
                else:
                    popup = WarningPopup()
                    popup.popup_text.text = "Looks like there are no EnergyPlus input files! Please visit the help tab to learn more."
                    popup.open()
        else:
            self.hvac_select.values = [hvac_option for hvac_option in hvac_options]

    def _get_location_options(self):
        """Get weather data from data bank."""
        try:
            data_manager = App.get_running_app().data_manager
            location_options = data_manager.data_bank['weather files'].keys()
        except AttributeError:
            pass
        except KeyError:
            if not os.path.exists(os.path.join(data_manager.data_bank_root, 'weather')):
                popup = WarningPopup()
                popup.popup_text.text = "Looks like there are no weather files downloaded!" \
                    + " Please visit the Data Manager to download data or the help tab to learn more."
                popup.open()
            else:
                data_root = os.path.join(data_manager.data_bank_root, 'weather')
                ls = [os.path.join(data_root, item) for item in os.listdir(data_root)]
                bool_ls = [True if os.path.isfile(item) else False for item in ls]

                if any(bool_ls):
                    bad_dir_popup = WarningPopup()
                    bad_dir_popup.popup_text.text = "There was an error processing the weather files." \
                        + " Please place your weather files in their own folder in the weather directory."
                    bad_dir_popup.open()
                else:
                    popup = WarningPopup()
                    popup.popup_text.text = "Looks like there are no weather files downloaded!" \
                        + " Please visit the Data Manager to download data or the help tab to learn more."
                    popup.open()
        else:
            self.location_select.values = [location_option for location_option in location_options]

    def _get_profile_options(self):
        """Get profile options from data bank and btm/valuation runs."""
        try:
            data_manager = App.get_running_app().data_manager
            profile_options = [key for key in data_manager.data_bank['profiles'].keys()] + [col for col in self.valuation_ops.columns] + \
                [col for col in self.btm_ops.columns]
        except AttributeError as e:
            logging.warning("Performance: {} A1".format(e))
            popup = WarningPopup()
            popup.popup_text.text = "Looks like there are no battery charge/discharge profiles! Please visit the help tab for more info."
            popup.open()
        except KeyError as e:
            logging.warning("Performance: {} Key1".format(e))
            if not os.path.exists(os.path.join(data_manager.data_bank_root, 'profile')):
                try:
                    profile_options = [col for col in self.valuation_ops.columns] + [col for col in self.btm_ops.columns]
                except AttributeError as e:
                    logging.warning("Performance: {} A2".format(e))
                    popup = WarningPopup()
                    popup.popup_text.text = "Looks like there are no battery charge/discharge profiles! Please visit the help tab for more info."
                    popup.open()
                else:
                    if profile_options:
                        self.profile_select.values = profile_options
                    else:
                        popup = WarningPopup()
                        popup.popup_text.text = "Looks like there are no battery charge/discharge profiles! Please visit the help tab for more info."
                        popup.open()
            else:
                data_root = os.path.join(data_manager.data_bank_root, 'profile')
                ls = [os.path.join(data_root, item) for item in os.listdir(data_root)]
                bool_ls = [True if os.path.isfile(item) else False for item in ls]

                if any(bool_ls):
                    bad_dir_popup = WarningPopup()
                    bad_dir_popup.popup_text.text = "There was an error processing the weather files." \
                        + " Please place your weather files in their own folder in the weather directory."
                    bad_dir_popup.open()
                else:
                    try:
                        profile_options = [col for col in self.valuation_ops.columns] + [col for col in self.btm_ops.columns]
                    except AttributeError as e:
                        logging.warning("Performance: {} A2".format(e))
                        popup = WarningPopup()
                        popup.popup_text.text = "Looks like there are no battery charge/discharge profiles! Please visit the help tab for more info."
                        popup.open()
                    else:
                        if profile_options:
                            self.profile_select.values = profile_options
                        else:
                            popup = WarningPopup()
                            popup.popup_text.text = "Looks like there are no battery charge/discharge profiles! Please visit the help tab for more info."
                            popup.open()
        else:
            if profile_options:
                self.profile_select.values = profile_options
            else:
                popup = WarningPopup()
                popup.popup_text.text = "Looks like there are no battery charge/discharge profiles! Please visit the help tab for more info."
                popup.open()

    def _reset_screen(self):

        # Deselects all RV selections.
        self.hvac_rv.deselect_all_nodes()
        self.location_rv.deselect_all_nodes()
        self.profile_rv.deselect_all_nodes()

        # Resets properties.
        self.hvac = ''
        self.location = ''
        self.profile = ''
        self.hvac_to_sim = None
        self.location_to_sim = None
        self.profile_to_sim = []

    def _hvac_selected(self):
        """Show input files."""
        self.hvac = self.hvac_select.text

        try:
            data_manager = App.get_running_app().data_manager

            self.hvac_rv.data = [{'name': idf_files[0], 'path':idf_files[1]}
                                 for idf_files in data_manager.data_bank['idf files'][self.hvac]]

        except AttributeError:
            pass
        else:
            Clock.schedule_once(partial(fade_in_animation, self.hvac_select_bx), 0)

    def _location_selected(self):
        """Show weather files."""
        self.location = self.location_select.text

        try:
            data_manager = App.get_running_app().data_manager

            self.location_rv.data = [{'name': location_files[0], 'path':location_files[1]}
                                     for location_files in data_manager.data_bank['weather files'][self.location] if '.epw' in location_files[0]]

        except AttributeError:
            pass
        else:
            Clock.schedule_once(partial(fade_in_animation, self.location_select_bx), 0)

    def _profile_selected(self):
        """Show selectable charge/discharge profiles."""
        self.profile = self.profile_select.text

        try:
            data_manager = App.get_running_app().data_manager
            profiles = data_manager.data_bank['profiles']
        except AttributeError:
            pass
        except KeyError:
            if self.profile in self.btm_ops.columns:
                self.profile_rv.data = [{'name': op['month'], 'path':None, 'op':op} for op in self.btm_ops[self.profile]]
            elif self.profile in self.valuation_ops.columns:
                self.profile_rv.data = [{'name': op['month'], 'path':None, 'op':op} for op in self.valuation_ops[self.profile]]
            else:
                print('The selected profile does not exist')

            Clock.schedule_once(partial(fade_in_animation, self.profile_select_bx), 0)
        else:
            if self.profile in profiles:
                self.profile_rv.data = [{'name': profile_files[0], 'path':profile_files[1], 'op':None}
                                        for profile_files in data_manager.data_bank['profiles'][self.profile]]
            elif self.profile in self.btm_ops.columns:
                self.profile_rv.data = [{'name': op['month'], 'path':None, 'op':op} for op in self.btm_ops[self.profile]]
            elif self.profile in self.valuation_ops.columns:
                self.profile_rv.data = [{'name': op['month'], 'path':None, 'op':op} for op in self.valuation_ops[self.profile]]
            else:
                print('The selected profile does not exist')

            Clock.schedule_once(partial(fade_in_animation, self.profile_select_bx), 0)

    def _validate_inputs(self):
        """Ensure proper selections are made."""
        if self.hvac_select.text == 'Select HVAC':
            popup = WarningPopup()
            popup.popup_text.text = 'Please select an HVAC.'
            popup.open()

        if self.location_select.text == 'Select location':
            popup = WarningPopup()
            popup.popup_text.text = 'Please select a location.'
            popup.open()

        if self.profile_select.text == 'Select profile':
            popup = WarningPopup()
            popup.popup_text.text = 'Please select a charge/discharge profile.'
            popup.open()

    def get_inputs(self):
        """Get data selections."""
        self._validate_inputs()

        hvac_selected = self.hvac_to_sim
        location_selected = self.location_to_sim
        profile_selected = self.profile_to_sim

        return hvac_selected, location_selected, profile_selected


class PerformanceSimParamScreen(Screen):
    """Screen to set parameters for performance simulations."""

    iso = StringProperty('')
    param_to_attr = dict()
    built = False

    def on_enter(self):
        """Build parameter widget."""
        if not self.built:
            self.built = self.param_widget.build()

    def _disable_text_input(self, text):
        # Disables the text input field of the parameter selected for a parameter sweep.
        for row_widget in self.param_widget.children:
            row_widget.text_input.disabled = False

        attr_name = self.param_to_attr.get(text, '')

        if attr_name:
            param_widget_row = getattr(self.param_widget, attr_name, None)
            param_widget_row.text_input.disabled = True

    def _validate_inputs(self):
        """Ensure inputs are made."""
        if len(self.param_widget.children) == 0:
            popup = WarningPopup()
            popup.popup_text.text = 'Please specify the simulation parameters.'
            popup.open()

    def get_inputs(self):
        """Get parameters."""
        self._validate_inputs()

        param_settings = self.param_widget.get_inputs(use_hint_text=True)

        return param_settings


class PerformanceSimSelectionsScreen(Screen):
    """Show the performance simulation data and parameter selections."""

    MODEL_PARAMS = {
        'eCap': 'Energy Capacity (MWh)',
        'pRat': 'Power Rating (MW)',
        'n_s': 'Self-Discharge Efficiency',
        'n_p': 'Power Electronics Efficiency',
        'q_rate': 'Battery Cell Ah Rating',
        'v_rate': 'Battery Cell Voltage Rating',
        'r': 'Battery Cell Internal Resistance',
        'k': 'Battery Cell k Parameter',
        'tau': 'Time Step (hr)',
        'h_setpoint': 'Heating Setpoint (C)',
        'c_setpoint': 'Cooling Setpoint (C)',
        'insulation': 'Insulation (m\u00b2 K/W)'
    }

    def __init__(self, **kwargs):
        super(PerformanceSimSelectionsScreen, self).__init__(**kwargs)

    def on_enter(self):
        """Show all of the selections."""
        Clock.schedule_once(partial(fade_in_animation, self.content), 0)

        data_screen = self.manager.get_screen('data')
        params_screen = self.manager.get_screen('params')

        hvac = data_screen.hvac
        location = data_screen.location
        profile = data_screen.profile
        hvac_to_sim = data_screen.hvac_to_sim['name']
        location_to_sim = data_screen.location_to_sim['name']
        profile_to_sim = data_screen.profile_to_sim
        param_inputs = params_screen.get_inputs()
        param_settings = [self.MODEL_PARAMS[param] + ': {}'.format(param_inputs[param]) for param in param_inputs]

        hvac_text = '\n    '.join(['[b]HVAC:[/b]',
                                   hvac,
                                   hvac_to_sim])
        self.hvac_label.text = hvac_text

        location_text = '\n    '.join(['[b]Location:[/b]',
                                       location,
                                       location_to_sim])
        self.location_label.text = location_text

        profile_text = '\n    '.join(['[b]Profile:[/b]',
                                      profile,
                                      ]+[profile['name'] for profile in profile_to_sim])
        self.profile_label.text = profile_text

        params_text = '[b]System Parameters:[/b]\n    '
        params_text += '\n    '.join(param_settings)
        self.params_label.text = params_text


class PerformanceSimCompletePopup(ValuationRunCompletePopup):
    """Popup for completed simulation."""

    def __init__(self, **kwargs):
        super(PerformanceSimCompletePopup, self).__init__(**kwargs)

        self.popup_text.text = 'Your performance simulation is completed.'

    def complete_open(self):
        """Open popup from main thread."""
        Clock.schedule_once(self.open)

    def complete_dismiss(self):
        """Close popup from main thread."""
        Clock.schedule_once(self.dismiss)


class PerformanceRunPopup(MyPopup):
    """Popup to ensure the user that the tool is running."""

    def __init__(self, **kwargs):
        super(PerformanceRunPopup, self).__init__(**kwargs)

        self.popup_text.text = 'Performance simulation is running. This may take a few minutes.'

    def run_open(self):
        """Open popup from main thread."""
        Clock.schedule_once(self.open)

    def run_dismiss(self):
        """Close popup from main thread."""
        Clock.schedule_once(self.dismiss)


class EnergyPlusPopup(WarningPopup):
    """Warning popup if EnergyPlus is missing."""

    def __init__(self, **kwargs):
        super(EnergyPlusPopup, self).__init__(**kwargs)

        self.popup_text.text = 'EnergyPlus not found. Please visit the help tab to configure Energyplus on your machine.'


class HVACEntry(RecycleViewRow):
    """HVAC entry for recycle view."""

    host_screen = None

    def apply_selection(self, rv, index, is_selected):
        """Respond to the selection of items in the view."""
        super(HVACEntry, self).apply_selection(rv, index, is_selected)

        if is_selected:
            self.host_screen.hvac_to_sim = rv.data[self.index]
        # else:
        #     self.host_screen.node = ''


class LocationEntry(RecycleViewRow):
    """Location entry for recycle view."""

    host_screen = None

    def apply_selection(self, rv, index, is_selected):
        """Respond to the selection of items in the view."""
        super(LocationEntry, self).apply_selection(rv, index, is_selected)

        if is_selected:
            self.host_screen.location_to_sim = rv.data[self.index]
        # else:
        #     self.host_screen.node = ''


class ProfileEntry(RecycleViewRow):
    """Profile Entry for recycle view."""

    host_screen = None

    def apply_selection(self, rv, index, is_selected):
        """Respond to the selection of items in the view."""
        super(ProfileEntry, self).apply_selection(rv, index, is_selected)

        if is_selected:
            if not rv.data[self.index] in self.host_screen.profile_to_sim:
                self.host_screen.profile_to_sim += [rv.data[self.index]]
        # else:
        #     self.host_screen.node = ''



from __future__ import absolute_import

from functools import partial

from kivy.app import App
from kivy.uix.screenmanager import Screen
from kivy.properties import StringProperty
from kivy.uix.gridlayout import GridLayout
from kivy.uix.textinput import TextInput

from es_gui.resources.widgets.common import WarningPopup, InputError, ValuationParameterRow, ValuationParameterWidget, ValuationRunCompletePopup
from es_gui.tools.valuation.valuation_optimizer import BadParameterException


class SetParametersScreen(Screen):
    """
    The screen for setting parameters for the energy storage model.
    """
    iso = StringProperty('')
    param_to_attr = dict()

    def on_pre_enter(self):
        self.iso = self.manager.get_screen('load_data').iso_select.text

    def on_iso(self, instance, value):
        while len(self.param_widget.children) > 0:
            for widget in self.param_widget.children:
                if isinstance(widget, ValuationParameterRow):
                    self.param_widget.remove_widget(widget)

        if self.iso:
            try:
                self.param_widget.build(self.iso)

                data_manager = App.get_running_app().data_manager
                MODEL_PARAMS = data_manager.get_valuation_model_params(value)

                self.param_to_attr = {param['name']: param['attr name']
                                      for param in MODEL_PARAMS}
                self.title.text = 'Set simulation parameters for the {market_area} market area.'.format(market_area=self.iso)
            except KeyError:
                pass

    def on_enter(self):
        # change the navigation bar title
        ab = self.manager.nav_bar
        ab.build_valuation_advanced_nav_bar()
        ab.set_title('Single Run: Set Parameters')

        data_manager = App.get_running_app().data_manager
        MODEL_PARAMS = data_manager.get_valuation_model_params(self.iso)

        if not MODEL_PARAMS:
            popup = WarningPopup()
            popup.bind(on_dismiss=partial(ab.go_to_screen, 'load_data'))
            popup.dismiss_button.text = 'Go back'
            popup.popup_text.text = 'We need a market area in the "Select Data" screen selected first to populate this area.'
            popup.open()
        # else:
        #     self.go_button.bind(on_release=self.manager.get_screen('valuation_advanced').open_valuation_run_menu)

    def _validate_inputs(self):
        self.param_widget.validate_inputs()

    def get_inputs(self):
        self.param_widget.get_inputs()
    
    def _generate_requests(self):
        data_screen = self.manager.get_screen('load_data')
        params_screen = self

        iso_selected, market_formulation_selected, node_selected, year_selected, month_selected = data_screen.get_inputs()
        param_settings = params_screen.get_inputs()

        requests = {'iso': iso_selected,
                    'market type': market_formulation_selected,
                    'months': [(month_selected, year_selected)],
                    'node id': node_selected,
                    'param set': param_settings
                    }

        return requests
    
    def execute_single_run(self, *args):
        try:
            requests = self._generate_requests()
        except ValueError as e:
            popup = WarningPopup()
            popup.popup_text.text = str(e)
            popup.open()
        except InputError as e:
            popup = WarningPopup()
            popup.popup_text.text = str(e)
            popup.open()
        except BadParameterException as e:
            popup = WarningPopup()
            popup.popup_text.text = str(e)
            popup.open()
        else:
            solver_name = App.get_running_app().config.get('optimization', 'solver')

            handler = self.manager.get_screen('valuation_home').handler
            handler.solver_name = solver_name

            try:
                _, handler_status = handler.process_requests(requests)
            except BadParameterException as e:
                popup = WarningPopup()
                popup.popup_text.text = str(e)
                popup.open()
            else:
                self.completion_popup = ValuationSingleRunCompletePopup()
                self.completion_popup.view_results_button.bind(on_release=self._go_to_view_results)

                if not handler_status:
                    self.completion_popup.title = "Oops!"
                    self.completion_popup.popup_text.text = "The optimization model had issues being built and/or solved. This is most likely due to bad data. No results have been recorded."

                self.completion_popup.open()            
    
    def _go_to_view_results(self, *args):
        self.manager.nav_bar.go_to_screen('valuation_results_viewer')
        self.completion_popup.dismiss()


class ValuationSingleRunCompletePopup(ValuationRunCompletePopup):
    def __init__(self, **kwargs):
        super(ValuationSingleRunCompletePopup, self).__init__(**kwargs)

        self.popup_text.text = 'Your specified valuation job has been completed.'



from __future__ import absolute_import
import logging
import os
import json

import pandas as pd

from es_gui.tools.dms import DataManagementSystem
from es_gui.tools.valuation.utilities import *


class ValuationDMS(DataManagementSystem):
    """
    A class for managing data for the energy storage valuation optimization functions. Class methods for each type of file to be loaded are included, extending from the get_data() method of the superclass. Each of these methods uses get_data() to retrieve the relevant data and loads the file and adds it to the DMS if the data is not loaded. An optional class method for calling each of the individual data methods can be included to, e.g., form the necessary arguments and return the desired variables.

    :param home_path: A string indicating the relative path to where data is saved.
    """
    def __init__(self, home_path, **kwargs):
        DataManagementSystem.__init__(self, **kwargs)

        self.home_path = home_path

        # with open(os.path.abspath(os.path.join(self.home_path, '..', 'es_gui', 'apps', 'valuation', 'definitions', 'nodes.json')), 'r') as fp:
        #     self.NODES = json.load(fp)

        #self.node_names = pd.read_excel(self.home_path+'nodeid.xlsx', sheetname=None)
        self.delimiter = ' @ '  # delimiter used to split information in id_key

    def get_node_name(self, node_id, ISO):
        """
        Retrieves the node name corresponding to the given node_id using the lookup table loaded during initialization.
        :param node_id: A string or int of a node ID.
        :return: The corresponding node name as a string.
        """
        # TODO: map node_id to node name

        return str(node_id)

        # try:
        #     node_name = self.NODES[ISO][node_id]['name']
        # except KeyError as e:
        #     raise e
        # else:
        #     return node_name

    def get_ercot_spp_data(self, id_key):
        """Retrieves DAM-SPP data for ERCOT."""
        logging.info('DMS: Loading ERCOT DA-SPP')
        try:
            # attempt to access data if it is already loaded
            spp_da = self.get_data(id_key)
        except KeyError:
            # load the data and add it to the DMS

            # deconstruct id_key to obtain args for read function
            spp_da = read_ercot_da_spp(*id_key.split(self.delimiter))
            self.add_data(spp_da, id_key)
        finally:
            return spp_da

    def get_ercot_ccp_data(self, id_key):
        """Retrieves DAM-CCP data for ERCOT."""
        logging.info('DMS: Loading ERCOT DA-CCP')
        try:
            # attempt to access data if it is already loaded
            REGUP = self.get_data(id_key + self.delimiter + 'REGUP')
            REGDN = self.get_data(id_key + self.delimiter + 'REGDN')
        except KeyError:
            # load the data and add it to the DMS

            # deconstruct id_key to obtain args for read function
            REGDN, REGUP = read_ercot_da_ccp(*id_key.split(self.delimiter)[:2])

            self.add_data(REGUP, id_key + self.delimiter + 'REGUP')
            self.add_data(REGDN, id_key + self.delimiter + 'REGDN')
        finally:
            return REGDN, REGUP

    def get_ercot_data(self, year, month, settlement_point):
        # construct file name paths
        path = os.path.join(self.home_path, 'ERCOT')  # path to data_bank root

        if isinstance(month, int):
            month = str(month)

        spp_fpath = os.path.join(path, 'SPP', str(year))
        
        try:
            for filename in os.listdir(spp_fpath):
                if filename.lower().endswith('.xlsx'):
                    fname = filename
        except ValueError as e:
            raise(e)

        spp_fname = os.path.join(spp_fpath, fname)

        ccp_fpath = os.path.join(path, 'CCP', str(year))

        try:
            for filename in os.listdir(ccp_fpath):
                if filename.lower().endswith('.csv'):
                    fname = filename
        except ValueError as e:
            raise(e)

        ccp_fname = os.path.join(ccp_fpath, fname)

        # construct identifier keys
        spp_id = self.delimiter.join([spp_fname, month, settlement_point])
        ccp_id = self.delimiter.join([ccp_fname, month])

        # retrieve data
        spp_da = self.get_ercot_spp_data(spp_id)
        rd, ru = self.get_ercot_ccp_data(ccp_id)

        return spp_da, rd, ru

    def get_pjm_lmp_data(self, *args):
        """Deprecated since 1.0"""
        logging.info('DMS: Loading PJM DA-LMP')
        try:
            # attempt to access data if it is already loaded
            lmp_da = self.get_data(*args)
        except KeyError:
            # load the data and add it to the DMS
            lmp_da = read_pjm_da_lmp(*args)
            self.add_data(lmp_da, *args)
        finally:
            return lmp_da

    def get_pjm_reg_price_data(self, *args):
        """Deprecated since 1.0"""
        logging.info('DMS: Loading PJM regulation prices')
        try:
            # attempt to access data if it is already loaded
            RegCCP = self.get_data(*args+('RegCCP',))
            RegPCP = self.get_data(*args+('RegPCP',))
        except KeyError:
            # load the data and add it to the DMS
            RegCCP, RegPCP = read_pjm_reg_price(*args)
            self.add_data({'RegCCP': RegCCP, 'RegPCP': RegPCP}, *args)
        finally:
            return RegCCP, RegPCP

    def get_pjm_mileage_data(self, *args):
        """Deprecated since 1.0"""
        logging.info('DMS: Loading PJM mileage data')
        try:
            # attempt to access data if it is already loaded
            MR = self.get_data(*args+('MR',))
            RA = self.get_data(*args+('RA',))
            RD = self.get_data(*args+('RD',))
        except KeyError:
            # load the data and add it to the DMS
            MR, RA, RD = read_pjm_mileage(*args)
            self.add_data({'MR': MR, 'RA': RA, 'RD': RD}, *args)
        finally:
            return MR, RA, RD

    def get_pjm_reg_signal_data(self, *args):
        """Deprecated since 1.0"""
        logging.info('DMS: Loading PJM regulation signal')
        try:
            # attempt to access data if it is already loaded
            RUP = self.get_data(*args+('RegUp',))
            RDW = self.get_data(*args+('RegDown',))
        except KeyError:
            # load the data and add it to the DMS
            RUP, RDW = read_pjm_reg_signal(*args)
            self.add_data({'RegUp': RUP, 'RegDown': RDW}, *args)
        finally:
            return RUP, RDW

    def get_pjm_data(self, year, month, nodeid):
        path = os.path.join(self.home_path, 'PJM')
        
        nodeid = str(nodeid)
        year = str(year)
        month = str(month)

        lmp_key = self.delimiter.join([path, year, month, nodeid, 'SPP'])
        mr_key = self.delimiter.join([path, year, month, 'MR'])
        ra_key = self.delimiter.join([path, year, month, 'RA'])
        rd_key = self.delimiter.join([path, year, month, 'RD'])
        rccp_key = self.delimiter.join([path, year, month, 'RegCCP'])
        rpcp_key = self.delimiter.join([path, year, month, 'RegPCP'])

        try:
            # attempt to access data if it is already loaded
            lmp_da = self.get_data(lmp_key)
            MR = self.get_data(mr_key)
            RA = self.get_data(ra_key)
            RD = self.get_data(rd_key)
            RegCCP = self.get_data(rccp_key)
            RegPCP = self.get_data(rpcp_key)
        except KeyError:
            # load the data and add it to the DMS
            lmp_da, MR, RA, RD, RegCCP, RegPCP = read_pjm_data(path, year, month, nodeid)

            self.add_data(lmp_da, lmp_key)
            self.add_data(MR, mr_key)
            self.add_data(RA, ra_key)
            self.add_data(RD, rd_key)
            self.add_data(RegCCP, rccp_key)
            self.add_data(RegPCP, rpcp_key)

        return lmp_da, MR, RA, RD, RegCCP, RegPCP
    
    def get_miso_lmp_data(self, *args):
        """Deprecated since 1.0"""
        logging.info('DMS: Loading MISO DA-LMP')
        try:
            # attempt to access data if is already loaded
            lmp_da = self.get_data(*args)
        except KeyError:
            # load the data and add it to the DMS
            lmp_da = read_miso_da_lmp(*args)
            self.add_data(lmp_da, *args)
        finally:
            return lmp_da

    def get_miso_reg_data(self, *args):
        """Deprecated since 1.0"""
        logging.info('DMS: Loading MISO RegMCP')
        try:
            # attempt to access data if is already loaded
            RegMCP = self.get_data(*args)
        except KeyError:
            # load the data and add it to the DMS
            RegMCP = read_miso_reg_price(*args)
            self.add_data(RegMCP, *args)
        finally:
            return RegMCP

    def get_miso_data(self, year, month, nodeid):
        path = os.path.join(self.home_path, 'MISO')

        year = str(year)
        month = str(month)

        lmp_key = self.delimiter.join([path, year, month, nodeid, 'LMP'])
        regmcp_key = self.delimiter.join([path, year, month, 'MCP'])

        try:
            # attempt to access data if it is already loaded
            lmp_da = self.get_data(lmp_key)
            RegMCP = self.get_data(regmcp_key)
        except KeyError:
            # load the data and add it to the DMS
            lmp_da, RegMCP = read_miso_data(path, year, month, nodeid)

            self.add_data(lmp_da, lmp_key)
            self.add_data(RegMCP, regmcp_key)

        return lmp_da, RegMCP

    ####################################################################################################################

    def get_isone_data(self, year, month, nodeid):
        path = os.path.join(self.home_path, 'ISONE')

        year = str(year)
        month = str(month)

        lmp_key = self.delimiter.join([path, year, month, nodeid, 'LMP'])
        rccp_key = self.delimiter.join([path, year, month, 'RegCCP'])
        rpcp_key = self.delimiter.join([path, year, month, 'RegPCP'])
        mimult_key = self.delimiter.join([path, year, month, 'MiMult'])

        try:
            # attempt to access data if it is already loaded
            lmp_da = self.get_data(lmp_key)
            rccp = self.get_data(rccp_key)
            rpcp = self.get_data(rpcp_key)
            mi_mult = self.get_data(mimult_key)
        except KeyError:
            # load the data and add it to the DMS
            lmp_da, rccp, rpcp, mi_mult = read_isone_data(path, year, month, nodeid)

            self.add_data(lmp_da, lmp_key)
            self.add_data(rccp, rccp_key)
            self.add_data(rpcp, rpcp_key)
            self.add_data(mi_mult, mimult_key)

        return lmp_da, rccp, rpcp, mi_mult

    ####################################################################################################################

    def get_nyiso_data(self, year, month, nodeid):
        path = os.path.join(self.home_path, 'NYISO')

        nodeid = str(nodeid)
        year = str(year)
        month = str(month)

        lbmp_key = self.delimiter.join([path, year, month, nodeid, 'LBMP'])
        rcap_key = self.delimiter.join([path, year, month, 'RegCAP'])

        try:
            # attempt to access data if it is already loaded
            lbmp_da = self.get_data(lbmp_key)
            rcap_da = self.get_data(rcap_key)
        except KeyError:
            # load the data and add it to the DMS
            lbmp_da, lbmp_rt, rcap_da, rcap_rt, rmov_da = read_nyiso_data(path, year, month, nodeid, typedat="both", RT_DAM="DAM")

            self.add_data(lbmp_da, lbmp_key)
            self.add_data(rcap_da, rcap_key)

        return lbmp_da, rcap_da


    def get_spp_data(self, year, month, nodeid):
        path = os.path.join(self.home_path, 'SPP')

        year = str(year)
        month = str(month)

        lmp_key = self.delimiter.join([path, year, month, nodeid, 'LMP'])
        mcpru_key = self.delimiter.join([path, year, month, 'MCPRU'])
        mcprd_key = self.delimiter.join([path, year, month, 'MCPRD'])

        try:
            # attempt to access data if it is already loaded
            lmp_da = self.get_data(lmp_key)
            mcpru_da = self.get_data(mcpru_key)
            mcprd_da = self.get_data(mcprd_key)
        except KeyError:
            # load the data and add it to the DMS
            # lmp_da, MR, RA, RD, RegCCP, RegPCP = read_pjm_data(path, year, month, nodeid)
            lmp_da, mcpru_da, mcprd_da = read_spp_data(path, year, month, nodeid, typedat="both")

            self.add_data(lmp_da, lmp_key)
            self.add_data(mcpru_da, mcpru_key)
            self.add_data(mcprd_da, mcprd_key)

        return lmp_da, mcpru_da, mcprd_da


    def get_caiso_data(self, year, month, nodeid):
        path = os.path.join(self.home_path, 'CAISO')

        year = str(year)
        month = str(month)

        lmp_key = self.delimiter.join([path, year, month, nodeid, 'LMP'])
        aspru_key = self.delimiter.join([path, year, month, 'ASPRU'])
        asprd_key = self.delimiter.join([path, year, month, 'ASPRD'])
        asprmu_key = self.delimiter.join([path, year, month, 'ASPRMU'])
        asprmd_key = self.delimiter.join([path, year, month, 'ASPRMD'])
        rmu_mm_key = self.delimiter.join([path, year, month, 'RMU_MM'])
        rmd_mm_key = self.delimiter.join([path, year, month, 'RMD_MM'])
        rmu_pacc_key = self.delimiter.join([path, year, month, 'RMU_PACC'])
        rmd_pacc_key = self.delimiter.join([path, year, month, 'RMD_PACC'])

        try:
            # attempt to access data if it is already loaded
            lmp_da = self.get_data(lmp_key)
            aspru_da = self.get_data(aspru_key)
            asprd_da = self.get_data(asprd_key)
            asprmu_da = self.get_data(asprmu_key)
            asprmd_da = self.get_data(asprmd_key)
            rmu_mm = self.get_data(rmu_mm_key)
            rmd_mm = self.get_data(rmd_mm_key)
            rmu_pacc = self.get_data(rmu_pacc_key)
            rmd_pacc = self.get_data(rmd_pacc_key)
        except KeyError:
            # load the data and add it to the DMS
            # lmp_da, MR, RA, RD, RegCCP, RegPCP = read_pjm_data(path, year, month, nodeid)
            lmp_da, aspru_da, asprd_da, asprmu_da, asprmd_da, rmu_mm, rmd_mm, rmu_pacc, rmd_pacc = read_caiso_data(path, year, month, nodeid)

            self.add_data(lmp_da, lmp_key)
            self.add_data(aspru_da, aspru_key)
            self.add_data(asprd_da, asprd_key)
            self.add_data(asprmu_da, asprmu_key)
            self.add_data(asprmd_da, asprmd_key)
            self.add_data(rmu_mm, rmu_mm_key)
            self.add_data(rmd_mm, rmd_mm_key)
            self.add_data(rmu_pacc, rmu_pacc_key)
            self.add_data(rmd_pacc, rmd_pacc_key)

        return lmp_da, aspru_da, asprd_da, asprmu_da, asprmd_da, rmu_mm, rmd_mm, rmu_pacc, rmd_pacc

    ####################################################################################################################

if __name__ == '__main__':
    dms = ValuationDMS(save_name='valuation_dms.p', home_path='data')
    
    # # ERCOT - data doesn't exist
    # year = 2010
    # month = 1
    # settlement_point = 'HB_HOUSTON'

    # spp_da, rd, ru = dms.get_ercot_data(year, month, settlement_point)

    # # ERCOT
    # year = 2010
    # month = 12
    # settlement_point = 'HB_HOUSTON'

    # spp_da, rd, ru = dms.get_ercot_data(year, month, settlement_point)

    # PJM
    # year = 2016
    # month = 5
    # nodeid = 1

    # lmp_da, MR, RA, RD, RegCCP, RegPCP = dms.get_pjm_data(year, month, nodeid)

    # MISO
    year = 2015
    month = 3
    nodeid = 'AEC'

    lmp_da, RegMCP = dms.get_miso_data(year, month, nodeid)


from __future__ import absolute_import

from functools import partial
from threading import Thread
import os
import calendar

from kivy.app import App
from kivy.uix.screenmanager import Screen
from kivy.uix.dropdown import DropDown
from kivy.uix.modalview import ModalView
from kivy.properties import DictProperty

from es_gui.apps.data_manager.data_manager import DataManagerException
from es_gui.tools.valuation.utilities import *
from es_gui.resources.widgets.common import MyPopup, InputError, RecycleViewRow


class LoadDataScreen(Screen):
    """The screen for loading data for the energy valuation application."""
    # loading = False
    # thread_list = []
    node = DictProperty()

    def __init__(self, **kwargs):
        super(LoadDataScreen, self).__init__(**kwargs)

        SingleRunNodeRecycleViewRow.host_screen = self
        NodeSelectMenu.host_screen = self
        self.node_select_menu = NodeSelectMenu()

    def on_enter(self):
        self.load_toolbar()

        # change the navigation bar title
        ab = self.manager.nav_bar
        ab.build_valuation_advanced_nav_bar()
        ab.set_title('Single Run: Select Data')
    
    def on_node(self, instance, value):
        try:
            logging.info('ValuationSingle: Pricing node changed to {0}.'.format(value['name']))
        except KeyError:
            logging.info('ValuationSingle: Pricing node selection reset.')
        else:
            self.node_select.text = value['name']
        finally:
            self.type_select.text = 'Select revenue streams'
            self.year_select.text = 'Select year'
            self.month_select.text = 'Select month'

    def _validate_inputs(self):
        iso_selected = self.iso_select.text

        data_manager = App.get_running_app().data_manager

        try:
            self.iso_select.values.index(iso_selected)
        except ValueError:
            raise InputError('Select a market area in "Select Data."')

        try:
            node_selected = self.node['nodeid']
            market_models_available = data_manager.get_valuation_revstreams(iso_selected, node_selected)
        except KeyError:
            raise InputError('Select a pricing node in "Select Data."')

        if node_selected not in data_manager.get_nodes(iso_selected).keys():
            raise InputError('Select a pricing node in "Select Data."')

        rev_streams = self.type_select.text

        if rev_streams not in market_models_available.keys():
            raise InputError('Select revenue streams in "Select Data."')

        try:
            year_selected = int(self.year_select.text)
        except ValueError:
            raise InputError('Select a year in "Select Data."')

        try:
            month_selected = list(calendar.month_name).index(self.month_select.text)
        except ValueError:
            raise InputError('Select a month in "Select Data."')

    def get_inputs(self):
        self._validate_inputs()

        iso_selected = self.iso_select.text

        data_manager = App.get_running_app().data_manager
        market_models_available = data_manager.get_valuation_revstreams(self.iso_select.text, self.node['nodeid'])
        market_formulation_selected = market_models_available[self.type_select.text]['market type']

        #node_selected = [self.node_select_menu.node_rv.data[selected_ix] for selected_ix in self.node_select_menu.node_rv.layout_manager.selected_nodes][0]

        node_selected = self.node['nodeid']

        year_selected = self.year_select.text
        month_selected = str(list(calendar.month_name).index(self.month_select.text))

        return iso_selected, market_formulation_selected, node_selected, year_selected, month_selected

    def load_toolbar(self, **kwargs):
        data_manager = App.get_running_app().data_manager
        iso_options = data_manager.get_markets()
        
        self.iso_select.values = iso_options
        self.type_select.values = []
        self.year_select.values = []
        self.month_select.values = []

    def reset_toolbar(self, **kwargs):
        self.type_select.text = 'Select revenue streams'
        self.year_select.text = 'Select year'
        self.month_select.text = 'Select month'
        self.node_select.text = 'Select node'

        if self.iso_select.text=='PJM':
            self.iso_img.source=os.path.join('es_gui', 'resources', 'images', 'IRCmap_pjm.png')
        elif self.iso_select.text=='MISO':
            self.iso_img.source=os.path.join('es_gui', 'resources', 'images', 'IRCmap_miso.png')
        elif self.iso_select.text=='ERCOT':
            self.iso_img.source=os.path.join('es_gui', 'resources', 'images', 'IRCmap_ercot.png')
        elif self.iso_select.text=='ISONE':
            self.iso_img.source=os.path.join('es_gui', 'resources', 'images', 'IRCmap_isone.png')
        elif self.iso_select.text=='NYISO':
            self.iso_img.source=os.path.join('es_gui', 'resources', 'images', 'IRCmap_nyiso.png')
        elif self.iso_select.text=='CAISO':
            self.iso_img.source=os.path.join('es_gui', 'resources', 'images', 'IRCmap_caiso.png')
        elif self.iso_select.text=='SPP':
            self.iso_img.source=os.path.join('es_gui', 'resources', 'images', 'IRCmap_spp.png')

        data_manager = App.get_running_app().data_manager
        node_data = [{'name': node[1],
                      'nodeid': node[0]} for node in data_manager.get_nodes(self.iso_select.text).items()]

        self.node_select_menu.node_rv.data = node_data
        self.node_select_menu.node_rv.unfiltered_data = node_data
        self.node_select_menu.node_rv.deselect_all_nodes()
        self.node = {}

    def load_year(self, iso):
        try:
            data_manager = App.get_running_app().data_manager
            year_options = data_manager.get_historical_data_options(self.iso_select.text, self.node['nodeid'], self.type_select.text).keys()
            self.year_select.values = year_options
        except (KeyError, DataManagerException):
            self.year_select.values = []

    def load_month(self, iso, year):
        try:
            data_manager = App.get_running_app().data_manager
            hist_data_options = data_manager.get_historical_data_options(self.iso_select.text, self.node['nodeid'], self.type_select.text)
            month_options = hist_data_options[str(year)]
            self.month_select.values = [calendar.month_name[int(month_index)] for month_index in month_options]
            
        except (KeyError, DataManagerException):
            self.month_select.values = []

    def load_nodeid(self, iso):
        try:
            self.node_select_menu.open()
        except KeyError:
            pass

    def load_dtype(self, iso):
        try:
            data_manager = App.get_running_app().data_manager
            market_models_available = data_manager.get_valuation_revstreams(self.iso_select.text, self.node['nodeid'])
            self.type_select.values = market_models_available.keys()
        except (KeyError, DataManagerException):
            self.type_select.values = []
    
    def enable_load(self):
        self.load_button.disabled=False

    # def load_data_function(self, iso, dttype, year, month, nodeid, op=None):
    #     mes_pop=MessagePopup()
    #     try:
    #         path = self.dpath[iso]
    #         flag = True
    #     except KeyError as e:
    #         flag = False
    #         mes_pop.title='Loading Data Error!'
    #         mes_pop.pop_label.text='Invalid ISO!'
    #         mes_pop.open()
    #         print(repr(e))
    #         raise(KeyError('Invalid ISO specified.'))

    #     try:
    #         yeari = int(year)
    #         monthi = int(month)
    #     except ValueError as e:
    #         flag = False
    #         mes_pop.title='Loading Data Error!'
    #         mes_pop.pop_label.text='Invalid year or month!'
    #         mes_pop.open()
    #         print(repr(e))
    #         raise(ValueError('Invalid year or month specified.'))


    #     if op is None:
    #         op_screen = self.manager.get_screen('valuation_advanced')
    #         op = op_screen.op

    #     dms = self.manager.get_screen('valuation_home').dms

    #     if flag and iso == 'PJM':
    #         node_name = dms.get_node_name(nodeid, iso)

    #         lmp_da, RUP, RDW, MR, RA, RD, RegCCP, RegPCP = dms.get_pjm_data(yeari, monthi, node_name)

    #         if dttype == 'Arbitrage and regulation':
    #             op.market_type = 'pjm_pfp'
    #             op.price_electricity = lmp_da
    #             op.mileage_ratio = MR
    #             op.mileage_slow = RA
    #             op.mileage_fast = RD
    #             op.price_reg_capacity = RegCCP
    #             op.price_reg_performance = RegPCP
    #             op.fraction_reg_up = RUP
    #             op.fraction_reg_down = RDW

    #         print('Finished loading data!')
    #     elif flag and iso == 'ERCOT':
    #         node_name = dms.get_node_name(nodeid, iso)

    #         lmp_da, rd, ru = dms.get_ercot_data(yeari, monthi, node_name)

    #         if dttype == 'Arbitrage and regulation':
    #             op.market_type = 'ercot_arbreg'
    #             op.price_electricity = lmp_da
    #             op.price_reg_up = ru
    #             op.price_reg_down = rd

    #         print('Finished loading data!')
    #     elif flag and iso == 'MISO':
    #         node_name = dms.get_node_name(nodeid, iso)

    #         lmp_da, RegMCP = dms.get_miso_data(yeari, monthi, node_name)
            
    #         if dttype == 'Arbitrage and regulation':
    #             op.market_type = 'miso_pfp'
    #             op.price_electricity = lmp_da
    #             op.price_reg_performance = RegMCP

    #         print('Finished loading data!')

    #     elif flag == True and iso == 'ISO-NE':
    #         node_name = dms.get_node_name(nodeid, iso)
    #         # nodeidi = int(nodeid)
    #         daLMP, RegCCP, RegPCP = dms.get_isone_data(year, month, nodeid)

    #         if dttype == 'Arbitrage and regulation':
    #             op.market_type = 'isone_pfp'
    #             op.price_electricity = daLMP
    #             op.price_reg_capacity = RegCCP
    #             op.price_reg_performance = RegPCP

    #         print('Finished loading data!')

    # def load_data(self, iso, dttype, year, month, nodeid):
    #     for th in self.thread_list:
    #         if th.isAlive():
    #             self.loading=True
    #         else:
    #             self.loading=False

    #     if not self.loading:
    #         t=Thread(target=self.load_data_function,args=(iso,dttype,year,month,nodeid))
    #         self.thread_list.append(t)
    #         t.start()
    #     else:
    #         mes_pop=MessagePopup()
    #         mes_pop.title='Loading Data Status'
    #         mes_pop.pop_label.text='Loading data is currently running!'
    #         mes_pop.open()
    #         print('Loading data is currently running')
    #         #print self.thread_list
    #         #print self.loading


class MessagePopup(MyPopup):
    pass


class NodeSelectMenu(ModalView):
    host_screen = None

    def on_dismiss(self):
        try:
            node_selected = [self.node_rv.data[selected_ix] for selected_ix in self.node_rv.layout_manager.selected_nodes][0]
            self.host_screen.node = node_selected
            #self.host_screen.node_select.text = node_selected['name']
        except IndexError:
            self.host_screen.node = {}
            #self.host_screen.node_select.text = 'Select node'


class SingleRunNodeRecycleViewRow(RecycleViewRow):
    """The representation widget for node in the node selector RecycleView."""
    host_screen = None

    def on_touch_down(self, touch):
        """Add selection on touch down."""
        if super(SingleRunNodeRecycleViewRow, self).on_touch_down(touch):
            return True
        if self.collide_point(*touch.pos) and self.selectable:
            return self.parent.select_with_touch(self.index, touch)



import numpy as np


def pdSeriesIdxWhereTrue(pdSeries):
    """
    Return the indexes of a pandas series for which the corresponding values are True (boolean).

    Parameters
    ----------
    pdSeries: pandas series
        Input pandas series of the type Boolean

    Returns
    -------
    List
        Indexes of the pandas series that correspond to True values
    """
    
    return pdSeries.index[pdSeries].tolist()

    
def geom_mean(vals, weights=None):
    
    if weights is None:
        weights = np.ones(vals.shape[1])
        
    gmean = np.prod(vals**weights, axis=1) ** (1/np.sum(weights))
    
    return gmean


from __future__ import absolute_import

import os
from functools import partial

from kivy.uix.screenmanager import Screen
from kivy.app import App

# from es_gui.tools.valuation.valuation_dms import ValuationDMS
from es_gui.resources.widgets.common import WarningPopup, NavigationButton
from es_gui.tools.btm.btm_dms import BtmDMS
from es_gui.proving_grounds.help_carousel import HelpCarouselModalView
from .op_handler import BtmOptimizerHandler


class BehindTheMeterHomeScreen(Screen):
    """
    The home screen for doing behind-the-meter energy storage analysis.
    """
    def __init__(self, **kwargs):
        super(BehindTheMeterHomeScreen, self).__init__(**kwargs)

        # Initialize data management system.
        self.dms = BtmDMS(
            max_memory=App.get_running_app().config.getint('btm', 'btm_dms_size')*1000,
            save_data=bool(App.get_running_app().config.getint('btm', 'btm_dms_save')),
            save_name='btm_dms.p',
            home_path='data',
            )
        self.handler = BtmOptimizerHandler(App.get_running_app().config.get('optimization', 'solver'))
        self.handler.dms = self.dms

    def on_enter(self):
        ab = self.manager.nav_bar
        ab.reset_nav_bar()
        ab.set_title('Behind-the-Meter Applications')

        help_button = NavigationButton(
            text='help',
            on_release=self.open_help_carousel,
        )

        ab.action_view.add_widget(help_button)

        # data_manager = App.get_running_app().data_manager
        
        # # Check if any data is available.
        # if not data_manager.data_bank:
        #     no_data_popup = WarningPopup()
        #     no_data_popup.popup_text.text = "Looks like you haven't downloaded any data yet. Try using QuESt Data Manager to get some data before returning here!"
        #     no_data_popup.dismiss_button.text = "Got it, take me back!"

        #     no_data_popup.bind(on_dismiss=partial(ab.go_to_screen, 'index'))
        #     no_data_popup.open()
    
    def open_help_carousel(self, *args):
        """
        """
        help_carousel_view = HelpCarouselModalView()
        help_carousel_view.title.text = "QuESt BTM"

        slide_01_text = "QuESt BTM is an application with tools for analyzing behind-the-meter energy storage use cases."

        slide_02_text = "The Time-of-Use Cost Savings wizard estimates the cost savings with behind-the-meter energy storage.\n\nYou will need the following data to use this tool:\n* Utility rate structure\n* Load profile (or import your own)\n\nYou may also add a co-located photovoltaic power profile or import your own."

        slide_03_text = "Upon completion of the wizard, you will be taken to the summary report screen. There a number of reports you can browse through that summarize different aspects of the simulation results. A brief synopsis of each component of the results including some key numbers.\n\nThe 'Generate report' button can be used to produce a document that summarizes the wizard run."

        slide_04_text = "This document includes your input selections, a primer on the mathematical model used, and all of the charts from the wizard summary reports.\n\nThe resulting HTML document and images are saved to the /results/*/report directory. You can view the report in a web browser."

        slide_05_text = "You can view simulation results in more detail using the Results Viewer tool. You can plot time series data and export simulation results for external processing."

        slides = [
            (os.path.join("es_gui", "resources", "help_views", "btm", "01.png"), slide_01_text),
            (os.path.join("es_gui", "resources", "help_views", "btm", "02.png"), slide_02_text),
            (os.path.join("es_gui", "resources", "help_views", "common", "wizard_report", "01.png"), slide_03_text),
            (os.path.join("es_gui", "resources", "help_views", "common", "wizard_report", "02.png"), slide_04_text),
            (os.path.join("es_gui", "resources", "help_views", "common", "results_viewer", "00.png"), slide_05_text),
        ]

        help_carousel_view.add_slides(slides)
        help_carousel_view.open()
